from pathlib import Path
from abc import ABC, abstractmethod

from PyQt5 import QtCore
from PyQt5.QtWidgets import QMainWindow, QFileDialog

from ...utils._docs import fill_doc
from ...stream_recorder import StreamRecorder


class _metaclass_ControlGUI(type(QMainWindow), type(ABC)):
    pass


@fill_doc
class _ControlGUI(QMainWindow, ABC, metaclass=_metaclass_ControlGUI):
    """
    Class representing a base controller GUI.

    Parameters
    ----------
    %(viewer_scope)s
    """

    @abstractmethod
    def __init__(self, scope):
        super().__init__()
        self._scope = scope

    @abstractmethod
    def _load_gui(self):
        """
        Load the GUI.
        """
        pass

    @abstractmethod
    def _init_backend(self):
        """
        Initialize the backend.
        """
        pass

    def closeEvent(self, event):
        """
        Event called when closing the GUI.
        """
        event.accept()

    # --------------------------------------------------------------------
    @abstractmethod
    def _connect_signals_to_slots(self):
        """
        Event handler. Connect QT signals to slots.
        """
        # Recording
        self._ui.pushButton_start_recording.clicked.connect(
            self.onClicked_pushButton_start_recording)
        self._ui.pushButton_stop_recording.clicked.connect(
            self.onClicked_pushButton_stop_recording)
        self._ui.pushButton_set_recording_dir.clicked.connect(
            self.onClicked_pushButton_set_recording_dir)

    @QtCore.pyqtSlot()
    def onClicked_pushButton_start_recording(self):
        record_dir = self._ui.lineEdit_recording_dir.text()
        self._recorder = StreamRecorder(
            record_dir, stream_name=self._scope.stream_name, fif_subdir=True,
            verbose=False)
        self._recorder.start(blocking=False)
        self._ui.pushButton_stop_recording.setEnabled(True)
        self._ui.pushButton_start_recording.setEnabled(False)
        self._ui.statusBar.showMessage(f"[Recording to '{record_dir}']")

    @QtCore.pyqtSlot()
    def onClicked_pushButton_stop_recording(self):
        if self._recorder.state.value == 1:
            self._recorder.stop()
            self._ui.pushButton_start_recording.setEnabled(True)
            self._ui.pushButton_stop_recording.setEnabled(False)
            self._ui.statusBar.showMessage("[Not recording]")

    @QtCore.pyqtSlot()
    def onClicked_pushButton_set_recording_dir(self):
        defaultPath = str(Path.home())
        path_name = QFileDialog.getExistingDirectory(
            caption="Choose the recording directory", directory=defaultPath)

        if path_name:
            self._ui.lineEdit_recording_dir.setText(path_name)
            self._ui.pushButton_start_recording.setEnabled(True)

    # --------------------------------------------------------------------
    @property
    def scope(self):
        """
        Measuring scope.
        """
        return self._scope

    @property
    @abstractmethod
    def backend(self):
        """
        Display backend.
        """
        pass

    @property
    @abstractmethod
    def ui(self):
        """
        Control UI.
        """
        pass



"""
Mock trigger.
"""

from ._trigger import _Trigger
from ..utils._checks import _check_type
from ..utils._docs import fill_doc, copy_doc
from ..utils._logs import logger


@fill_doc
class MockTrigger(_Trigger):
    """
    Mock trigger class.

    Parameters
    ----------
    %(trigger_verbose)s
    """

    def __init__(self, *, verbose: bool = True):
        super().__init__(verbose)

    @copy_doc(_Trigger.signal)
    def signal(self, value: int) -> bool:
        _check_type(value, ('int', ), item_name='value')
        self._set_data(value)
        super().signal(value)
        return True

    def _signal_off(self):
        """
        Reset trigger signal to ``0``.
        """
        self._set_data(0)

    @copy_doc(_Trigger._set_data)
    def _set_data(self, value: int):
        super()._set_data(value)
        logger.info('MOCK trigger set to %i', value)



import math
from pathlib import Path
from configparser import RawConfigParser

from PyQt5 import QtCore
from PyQt5.QtWidgets import QHeaderView, QTableWidgetItem

from ._control import _ControlGUI
from ._ui_control import UI_MainWindow
from ..backends.pyqtgraph import _BackendPyQtGraph
from ...utils._logs import logger
from ...utils._docs import fill_doc, copy_doc


@fill_doc
class ControlGUI_EEG(_ControlGUI):
    """
    Controller GUI for EEG LSL Stream.

    Parameters
    ----------
    %(viewer_scope)s
    """

    def __init__(self, scope):
        super().__init__(scope)
        config_file = 'settings_scope_eeg.ini'

        self._load_configuration(config_file)
        self._load_gui()
        self._init_backend()
        self._connect_signals_to_slots()
        self._set_configuration(config_file)

        self._backend.start_timer()

    def _load_gui(self):
        """
        Loads the UI created with QtCreator.
        """
        self._ui = UI_MainWindow(self)

        for yRange in self._yRanges:
            self._ui.comboBox_signal_yRange.addItem(str(yRange))
            logger.debug(f'y-scale option {yRange} added.')

        # Set table channels row/col
        self._nb_table_columns = 8 if self._scope.nb_channels > 64 else 4
        self._nb_table_rows = math.ceil(
            self._scope.nb_channels / self._nb_table_columns)
        self._ui.table_channels.setRowCount(self._nb_table_rows)
        self._ui.table_channels.setColumnCount(self._nb_table_columns)

        logger.debug(
            'Channel table set to %d row and %d col.' %
            (self._nb_table_rows, self._nb_table_columns))

        # Set table channels elements
        for idx in range(self._scope.nb_channels):
            row = idx // self._nb_table_columns
            col = idx % self._nb_table_columns
            self._ui.table_channels.setItem(
                row, col, QTableWidgetItem(idx))
            self._ui.table_channels.item(row, col).setTextAlignment(
                QtCore.Qt.AlignCenter)
            self._ui.table_channels.item(row, col).setText(
                self._scope.channels_labels[idx])
            logger.debug('Added channel %s' % self._scope.channels_labels[idx])

        # Table channels header
        self._ui.table_channels.verticalHeader().setSectionResizeMode(
            QHeaderView.Stretch)
        self._ui.table_channels.horizontalHeader().setSectionResizeMode(
            QHeaderView.Stretch)

        # Set position on the screen
        self.setGeometry(100, 100,
                         self.geometry().width(),
                         self.geometry().height())
        self.setFixedSize(self.geometry().width(),
                          self.geometry().height())
        self.show()  # Display

    def _load_configuration(self, file):
        """
        Load default configuration for the ranges.
        """
        path2settings_folder = Path(__file__).parent / 'settings'
        scope_settings = RawConfigParser(
            allow_no_value=True, inline_comment_prefixes=('#', ';'))
        scope_settings.read(str(path2settings_folder/file))

        # yRange (amplitude)
        self._yRanges = {'1uV': 1, '10uV': 10, '25uV': 25,
                         '50uV': 50, '100uV': 100, '250uV': 250,
                         '500uV': 500, '1mV': 1000, '2.5mV': 2500,
                         '10mV': 10000}
        try:
            self._yRange = float(scope_settings.get("plot", "yRange"))
            if self._yRange not in self._yRanges.values:
                raise ValueError
        except Exception:  # Default to 25 uV
            self._yRange = 25.

        # xRange (time)
        try:
            self._xRange = int(scope_settings.get("plot", "xRange"))
        except Exception:  # Default to 10s
            self._xRange = 10

    def _set_configuration(self, file):
        """
        Load and set a default configuration for the GUI.
        """
        path2settings_folder = Path(__file__).parent / 'settings'
        scope_settings = RawConfigParser(
            allow_no_value=True, inline_comment_prefixes=('#', ';'))
        scope_settings.read(str(path2settings_folder/file))

        # yRange
        init_idx = list(self._yRanges.values()).index(self._yRange)
        self._ui.comboBox_signal_yRange.setCurrentIndex(init_idx)

        # xRange
        try:
            self._ui.spinBox_signal_xRange.setValue(self._xRange)
        except Exception:  # 10s by default
            self._ui.spinBox_signal_xRange.setValue(10)

        # CAR
        try:
            self._ui.checkBox_car.setChecked(bool(
                scope_settings.get("filtering", "apply_car_filter")))
        except Exception:
            self._ui.checkBox_car.setChecked(False)

        # BP Filters
        self._ui.checkBox_bandpass.setChecked(True)
        self._ui.checkBox_bandpass.setEnabled(False)
        try:
            self._ui.doubleSpinBox_bandpass_low.setValue(float(
                scope_settings.get(
                    "filtering", "bandpass_cutoff_frequency").split(' ')[0]))
            self._ui.doubleSpinBox_bandpass_high.setValue(float(
                scope_settings.get(
                    "filtering", "bandpass_cutoff_frequency").split(' ')[1]))
        except Exception:
            # Default to [1, 40] Hz.
            self._ui.doubleSpinBox_bandpass_low.setValue(1.)
            self._ui.doubleSpinBox_bandpass_high.setValue(40.)

        self._ui.doubleSpinBox_bandpass_high.setMinimum(
            self._ui.doubleSpinBox_bandpass_low.value()+1)
        self._ui.doubleSpinBox_bandpass_low.setMaximum(
            self._ui.doubleSpinBox_bandpass_high.value()-1)
        self._scope.init_bandpass_filter(
            low=self._ui.doubleSpinBox_bandpass_low.value(),
            high=self._ui.doubleSpinBox_bandpass_high.value())

        # Trigger events
        try:
            self._ui.checkBox_show_LPT_trigger_events.setChecked(
                bool(scope_settings.get("plot", "show_LPT_events")))
        except Exception:
            self._ui.checkBox_show_LPT_trigger_events.setChecked(False)

        # Table channels
        for idx in range(self._scope.nb_channels):
            row = idx // self._nb_table_columns
            col = idx % self._nb_table_columns
            self._ui.table_channels.item(row, col).setSelected(True)

        # Status bar
        self._ui.statusBar.showMessage("[Not recording]")

    @copy_doc(_ControlGUI._init_backend)
    def _init_backend(self):
        geometry = (self.geometry().x()+self.width(), self.geometry().y(),
                    self.width()*2, self.height())
        self._backend = _BackendPyQtGraph(
            self._scope, geometry, self._xRange, self._yRange)

    # --------------------------------------------------------------------
    @copy_doc(_ControlGUI._connect_signals_to_slots)
    def _connect_signals_to_slots(self):
        super()._connect_signals_to_slots()
        # Scales
        self._ui.comboBox_signal_yRange.activated.connect(
            self.onActivated_comboBox_signal_yRange)
        self._ui.spinBox_signal_xRange.valueChanged.connect(
            self.onValueChanged_spinBox_signal_xRange)

        # CAR / Filters
        self._ui.checkBox_car.stateChanged.connect(
            self.onClicked_checkBox_car)
        self._ui.checkBox_bandpass.stateChanged.connect(
            self.onClicked_checkBox_bandpass)
        self._ui.doubleSpinBox_bandpass_low.valueChanged.connect(
            self.onValueChanged_doubleSpinBox_bandpass_low)
        self._ui.doubleSpinBox_bandpass_high.valueChanged.connect(
            self.onValueChanged_doubleSpinBox_bandpass_high)

        # Trigger events
        self._ui.checkBox_show_LPT_trigger_events.stateChanged.connect(
            self.onClicked_checkBox_show_LPT_trigger_events)

        # Channel table
        self._ui.table_channels.itemSelectionChanged.connect(
            self.onSelectionChanged_table_channels)

    @QtCore.pyqtSlot()
    def onActivated_comboBox_signal_yRange(self):
        self._yRange = float(list(self._yRanges.values())[
            self._ui.comboBox_signal_yRange.currentIndex()])
        self._backend.yRange = self._yRange
        logger.debug('y-range set to %d' % self._yRange)

    @QtCore.pyqtSlot()
    def onValueChanged_spinBox_signal_xRange(self):
        self._xRange = int(self._ui.spinBox_signal_xRange.value())
        self._backend.xRange = self._xRange
        logger.debug('x-range set to %d' % self._xRange)

    @QtCore.pyqtSlot()
    def onClicked_checkBox_car(self):
        self._scope.apply_car = self._ui.checkBox_car.isChecked()
        logger.debug('CAR checkbox: %s' % self._ui.checkBox_car.isChecked())

    @QtCore.pyqtSlot()
    def onClicked_checkBox_bandpass(self):
        self._scope.apply_bandpass = self._ui.checkBox_bandpass.isChecked()
        logger.debug(
            'BP checkbox: %s' % self._ui.checkBox_bandpass.isChecked())

    @QtCore.pyqtSlot()
    def onValueChanged_doubleSpinBox_bandpass_low(self):
        self._ui.doubleSpinBox_bandpass_high.setMinimum(
            self._ui.doubleSpinBox_bandpass_low.value()+1)
        self._scope.init_bandpass_filter(
            low=self._ui.doubleSpinBox_bandpass_low.value(),
            high=self._ui.doubleSpinBox_bandpass_high.value())
        logger.debug(
            'BP set to [%d, %d]' %
            (self._ui.doubleSpinBox_bandpass_low.value(),
             self._ui.doubleSpinBox_bandpass_high.value()))

    @QtCore.pyqtSlot()
    def onValueChanged_doubleSpinBox_bandpass_high(self):
        self._ui.doubleSpinBox_bandpass_low.setMaximum(
            self._ui.doubleSpinBox_bandpass_high.value()-1)
        self._scope.init_bandpass_filter(
            low=self._ui.doubleSpinBox_bandpass_low.value(),
            high=self._ui.doubleSpinBox_bandpass_high.value())
        logger.debug(
            'BP set to [%d, %d]' %
            (self._ui.doubleSpinBox_bandpass_low.value(),
             self._ui.doubleSpinBox_bandpass_high.value()))

    @QtCore.pyqtSlot()
    def onClicked_checkBox_show_LPT_trigger_events(self):
        self._backend.show_LPT_trigger_events = bool(
            self._ui.checkBox_show_LPT_trigger_events.isChecked())
        logger.debug(
            'LPT trigger checkbox: %s' %
            self._ui.checkBox_show_LPT_trigger_events.isChecked())

    @QtCore.pyqtSlot()
    def onSelectionChanged_table_channels(self):
        selected = self._ui.table_channels.selectedItems()
        self._scope.selected_channels = sorted([
            item.row()*self._nb_table_columns + item.column()
            for item in selected])
        self._backend.selected_channels = self._scope.selected_channels

    def closeEvent(self, event):
        """
        Event called when closing the _ScopeControllerUI window.
        """
        if self._ui.pushButton_stop_recording.isEnabled():
            self.onClicked_pushButton_stop_recording()
        self._backend.close()
        super().closeEvent(event)

    # --------------------------------------------------------------------
    @property
    @copy_doc(_ControlGUI.backend)
    def backend(self):
        return self._backend

    @property
    @copy_doc(_ControlGUI.ui)
    def ui(self):
        return self._ui

    @property
    def xRange(self):
        """
        Selected X range (time) [seconds].
        """
        return self._xRange

    @property
    def yRange(self):
        """
        Selected Y range (amplitude) [uV].
        """
        return self._yRange



# (not using import * for Python 2.5 support)
from .pylsl import IRREGULAR_RATE, DEDUCED_TIMESTAMP, FOREVER, cf_float32,\
    cf_double64, cf_string, cf_int32, cf_int16, cf_int8, cf_int64,\
    cf_undefined, protocol_version, library_version, library_info, local_clock,\
    proc_ALL, proc_none, proc_clocksync, proc_dejitter, proc_monotonize, proc_threadsafe,\
    StreamInfo, StreamOutlet, resolve_streams, resolve_byprop,\
    resolve_bypred, StreamInlet, XMLElement, ContinuousResolver,\
    TimeoutError, LostError, InvalidArgumentError, InternalError,\
    stream_info, stream_outlet, stream_inlet, xml_element,\
    timeout_error, lost_error, vectorf, vectord, vectorl,\
    vectori, vectors, vectorc, vectorstr, resolve_stream

from .version import __version__



"""
Function fetching file from URL.
Adapted from MNE:
    https://github.com/mne-tools/mne-python
Adapted from NILEARN:
    https://github.com/nilearn/nilearn
"""

import time
import hashlib
import shutil
from math import log
from pathlib import Path
from mne.utils.progressbar import ProgressBar
from urllib import parse, request
from urllib.error import HTTPError, URLError

from ..utils._logs import logger


def fetch_file(url, file_name, print_destination=True, resume=True,
               hash_=None, hash_type='md5', timeout=30.):
    """
    Load requested file, downloading it if needed or requested.

    Parameters
    ----------
    url : str
        The url of the file to be downloaded.
    file_name : str
        Name, along with the path, of where downloaded file will be saved.
    print_destination : bool
        If true, destination of where file was saved will be printed after
        download finishes.
    resume : bool
        If True, try to resume partially downloaded files.
    hash_ : str | None
        The hash of the file to check. If None, no checking is performed.
    hash_type : str
        The type of hashing to use such as 'md5' or 'sha1'.
    timeout : float
        The URL open timeout.
    """
    if hash_ is not None:
        if not isinstance(hash_, str):
            raise ValueError(
                'Bad hash value given, should be a string. Given: %s.'
                % type(hash_))
        if hash_type == 'md5' and len(hash_) != 32:
            raise ValueError(
                'Bad hash value given, should be a 32-character string:\n%s'
                % hash_)
        if hash_type == 'sha1' and len(hash_) != 40:
            raise ValueError(
                'Bad hash value given, should be a 40-character string:\n%s'
                % hash_)
        if hash_type not in ['md5', 'sha1']:
            raise ValueError(
                "Unsupported hash type %s.\nSupported: 'md5', 'sha1'."
                % hash_type)
    file_name = Path(file_name)
    temp_file_name = file_name.with_suffix(file_name.suffix+'.part')
    scheme = parse.urlparse(url).scheme
    if scheme not in ('http', 'https'):
        raise NotImplementedError('Cannot use scheme %r' % (scheme,))
    try:
        # Triage resume
        if not temp_file_name.exists():
            resume = False
        if resume:
            with open(temp_file_name, 'rb', buffering=0) as local_file:
                local_file.seek(0, 2)
                initial_size = local_file.tell()
            del local_file
        else:
            initial_size = 0
        _get_http(url, temp_file_name, initial_size, timeout)

        # check hash sum
        if hash_ is not None:
            logger.info('Verifying hash %s.' % (hash_,))
            hashsum = _hashfunc(temp_file_name, hash_type=hash_type)
            if hash_ != hashsum:
                raise RuntimeError('Hash mismatch for downloaded file %s, '
                                   'expected %s but got %s'
                                   % (temp_file_name, hash_, hashsum))
        shutil.move(temp_file_name, file_name)
        if print_destination is True:
            logger.info('File saved as %s.\n' % file_name)
    except Exception:
        logger.error('Error while fetching file %s.'
                     ' Dataset fetching aborted.' % url)
        raise


def _hashfunc(fname, block_size=1048576, hash_type="md5"):
    """
    Calculate the hash for a file.

    Parameters
    ----------
    fname : str
        Filename.
    block_size : int
        Block size to use when reading.

    Returns
    -------
    hash_ : str
        The hexadecimal digest of the hash.
    """

    if hash_type == "md5":
        hasher = hashlib.md5()
    elif hash_type == "sha1":
        hasher = hashlib.sha1()
    else:
        raise ValueError("Unsupported hash type. Supported: 'md5', 'sha1'.")
    with open(fname, 'rb') as fid:
        while True:
            data = fid.read(block_size)
            if not data:
                break
            hasher.update(data)
    return hasher.hexdigest()


def _get_http(url, temp_file_name, initial_size, timeout):
    """Safely (resume a) download to a file from http(s)."""
    response = None
    extra = ''
    if initial_size > 0:
        logger.debug('  Resuming at %s' % (initial_size,))
        req = request.Request(
            url, headers={'Range': 'bytes=%s-' % (initial_size,)})
        try:
            response = request.urlopen(req, timeout=timeout)
            content_range = response.info().get('Content-Range', None)
            if (content_range is None or not content_range.startswith(
                    'bytes %s-' % (initial_size,))):
                raise IOError('Server does not support resuming')
        except (KeyError, HTTPError, URLError, IOError):
            initial_size = 0
            response = None
        else:
            extra = ', resuming at %s' % (_sizeof_fmt(initial_size),)
    if response is None:
        response = request.urlopen(request.Request(url), timeout=timeout)
    file_size = int(response.headers.get('Content-Length', '0').strip())
    file_size += initial_size
    url = response.geturl()
    logger.info('Downloading %s (%s%s)' % (url, _sizeof_fmt(file_size), extra))
    del url
    mode = 'ab' if initial_size > 0 else 'wb'
    progress = ProgressBar(file_size, initial_size, unit='B',
                           mesg='Downloading', unit_scale=True,
                           unit_divisor=1024)
    del file_size
    chunk_size = 8192  # 2 ** 13
    with open(temp_file_name, mode) as local_file:
        while True:
            t0 = time.time()
            chunk = response.read(chunk_size)
            dt = time.time() - t0
            if dt < 0.01:
                chunk_size *= 2
            elif dt > 0.1 and chunk_size > 8192:
                chunk_size = chunk_size // 2
            if not chunk:
                break
            local_file.write(chunk)
            progress.update_with_increment_value(len(chunk))


def _sizeof_fmt(num):
    """
    Turn number of bytes into human-readable str.

    Parameters
    ----------
    num : int
        The number of bytes.

    Returns
    -------
    size : str
        The size in human-readable format.
    """

    units = ['bytes', 'kB', 'MB', 'GB', 'TB', 'PB']
    decimals = [0, 0, 1, 2, 2, 2]
    if num > 1:
        exponent = min(int(log(num, 1024)), len(units) - 1)
        quotient = float(num) / 1024 ** exponent
        unit = units[exponent]
        num_decimals = decimals[exponent]
        format_string = '{0:.%sf} {1}' % (num_decimals)
        return format_string.format(quotient, unit)
    if num == 0:
        return '0 bytes'
    if num == 1:
        return '1 byte'



import sys
import time

from PyQt5.QtWidgets import QApplication

from .scope.scope_eeg import ScopeEEG
from .control_gui.control_eeg import ControlGUI_EEG
from ..stream_receiver import StreamReceiver, StreamEEG
from ..utils._logs import logger
from ..utils.lsl import search_lsl
from ..utils._checks import _check_type


class StreamViewer:
    """
    Class for visualizing the signals coming from an LSL stream. The stream
    viewer will connect to only one LSL stream. If ``stream_name`` is set to
    ``None``, an automatic search is performed followed by a prompt if multiple
    non-markers streams are found.

    Parameters
    ----------
    stream_name : str | None
        Servers' name to connect to. ``None`` will prompt the user.
    """

    def __init__(self, stream_name=None):
        self._stream_name = StreamViewer._check_stream_name(stream_name)

    def start(self, bufsize=0.2):
        """
        Connect to the selected amplifier and plot the streamed data.

        If ``stream_name`` is not provided, look for available streams on the
        network.

        Parameters
        ----------
        bufsize : int | float
            Buffer/window size of the attached StreamReceiver.
            The default ``0.2`` should work in most cases since data is fetched
            every 20 ms.
        """
        logger.info(f'Connecting to the stream: {self.stream_name}')
        self._sr = StreamReceiver(bufsize=bufsize, winsize=bufsize,
                                  stream_name=self._stream_name)
        self._sr.streams[self._stream_name].blocking = False
        time.sleep(bufsize)  # Delay to fill the LSL buffer.

        if isinstance(self._sr.streams[self._stream_name], StreamEEG):
            self._scope = ScopeEEG(self._sr, self._stream_name)
            app = QApplication(sys.argv)
            self._ui = ControlGUI_EEG(self._scope)
            sys.exit(app.exec_())
        else:
            logger.error(
                'Unsupported stream type '
                f'{type(self._sr.streams[self._stream_name])}')

    # --------------------------------------------------------------------
    @staticmethod
    def _check_stream_name(stream_name):
        """
        Checks that the stream_name is valid or search for a valid stream on
        the network.
        """
        _check_type(stream_name, (None, str), item_name='stream_name')
        if stream_name is None:
            stream_name = search_lsl(ignore_markers=True)
            if stream_name is None:
                raise RuntimeError('No LSL stream found.')
        return stream_name

    # --------------------------------------------------------------------
    @property
    def stream_name(self):
        """
        Connected stream's name.

        :type: str
        """
        return self._stream_name

    @property
    def sr(self):
        """
        Connected StreamReceiver.

        :type: StreamReceiver
        """
        return self._sr



from .mock import MockTrigger  # noqa: F401
from .parallel import ParallelPortTrigger  # noqa: F401
from .software import SoftwareTrigger  # noqa: F401
from .trigger_def import TriggerDef  # noqa: F401


__all__ = ['MockTrigger',
           'ParallelPortTrigger',
           'SoftwareTrigger',
           'TriggerDef']



import os
import sys
import glob

from pathlib import Path
from importlib import import_module

import bsl


def run():
    """Entrypoint for bsl <command> usage."""
    bsl_root = Path(__file__).parent.parent
    valid_commands = sorted(
        glob.glob(str(bsl_root/'commands'/'bsl_*.py')))
    valid_commands = [file.split(os.path.sep)[-1][4:-3]
                      for file in valid_commands]

    def print_help():
        print("Usage: BrainStreamingLayer command options\n")
        print("Accepted commands:\n")
        for command in valid_commands:
            print("\t- %s" % command)
        print('\nExample: bsl stream_player StreamPlayer "path to .fif file"')
        print("\nGetting help example: bsl stream_recorder -h")

    if len(sys.argv) == 1 or "help" in sys.argv[1] or "-h" in sys.argv[1]:
        print_help()
    elif sys.argv[1] == "--version":
        print("BrainStreamingLayer %s" % bsl.__version__)
    elif sys.argv[1] not in valid_commands:
        print('Invalid command: "%s"\n' % sys.argv[1])
        print_help()
    else:
        cmd = sys.argv[1]
        cmd = import_module('.bsl_%s' % (cmd,), 'bsl.commands')
        sys.argv = sys.argv[1:]
        cmd.run()



"""
BSL's logger.
"""
import sys
import logging


logger = logging.getLogger('bsl')
logger.propagate = False  # don't propagate (in case of multiple imports)


def init_logger(verbosity='INFO'):
    """
    Initialize a logger. Assign sys.stdout as a handler of the logger.

    Parameters
    ----------
    verbosity : int | str
        Logger verbosity.
    """
    set_log_level(verbosity)
    add_stream_handler(sys.stdout, verbosity)


def add_stream_handler(stream, verbosity='INFO'):
    """
    Add a handler to the logger. The handler redirects the logger output to
    the stream.

    Parameters
    ----------
    stream : The output stream, e.g. sys.stdout
    verbosity : int | str
        Handler verbosity.
    """
    handler = logging.StreamHandler(stream)
    handler.setFormatter(BSLformatter())
    logger.addHandler(handler)
    set_handler_log_level(verbosity, -1)


def add_file_handler(fname, mode='a', verbosity='INFO'):
    """
    Add a file handler to the logger. The handler saves the logs to file.

    Parameters
    ----------
    fname : str | Path
    mode : str
        Mode in which the file is openned.
    verbosity : int | str
        Handler verbosity.
    """
    handler = logging.FileHandler(fname, mode)
    handler.setFormatter(BSLformatter())
    logger.addHandler(handler)
    set_handler_log_level(verbosity, -1)


def set_handler_log_level(verbosity, handler_id=0):
    """
    Set the log level for a specific handler.
    First handler (ID 0) is always stdout, followed by user-defined handlers.

    Parameters
    ----------
    verbosity : int | str
        Logger verbosity.
    handler_id : int
        ID of the handler among 'logger.handlers'.
    """
    logger.handlers[handler_id].setLevel = verbosity


def set_log_level(verbosity):
    """
    Set the log level for the logger.

    Parameters
    ----------
    verbosity : int | str
        Logger verbosity.
    """
    logger.setLevel(verbosity)


class BSLformatter(logging.Formatter):
    """
    Format string Syntax for BSL.
    """
    # Format string syntax for the different Log levels
    _formatters = dict()
    _formatters[logging.DEBUG] = logging.Formatter(
        fmt="[%(module)s:%(funcName)s:%(lineno)d] %(levelname)s: "
            "%(message)s (%(asctime)s)")
    _formatters[logging.INFO] = logging.Formatter(
        fmt="[%(module)s.%(funcName)s] %(levelname)s: %(message)s")
    _formatters[logging.WARNING] = logging.Formatter(
        fmt="[%(module)s.%(funcName)s] %(levelname)s: %(message)s")
    _formatters[logging.ERROR] = logging.Formatter(
        fmt="[%(module)s:%(funcName)s:%(lineno)d] %(levelname)s: %(message)s")

    def __init__(self):
        super().__init__(fmt='%(levelname): %(message)s')

    def format(self, record):
        """
        Format the received log record.

        Parameters
        ----------
        record : logging.LogRecord
        """
        if record.levelno <= logging.DEBUG:
            return self._formatters[logging.DEBUG].format(record)
        elif record.levelno <= logging.INFO:
            return self._formatters[logging.INFO].format(record)
        elif record.levelno <= logging.WARNING:
            return self._formatters[logging.WARNING].format(record)
        else:
            return self._formatters[logging.ERROR].format(record)

        return super().format(record)


init_logger()



import copy
from abc import ABC, abstractmethod

import math

from ...utils._docs import fill_doc


@fill_doc
class _Backend(ABC):
    """
    Class representing a base backend.

    Parameters
    ----------
    %(viewer_scope)s
    %(viewer_backend_geometry)s
    %(viewer_backend_xRange)s
    %(viewer_backend_yRange)s
    """

    @abstractmethod
    def __init__(self, scope, geometry, xRange, yRange):
        self._scope = scope

        # Variables
        self._xRange = xRange  # duration in seconds
        self._yRange = yRange  # amplitude range in uV

        self._show_LPT_trigger_events = False
        self._selected_channels = copy.deepcopy(self._scope.selected_channels)

    def _init_variables(self):
        """
        Initialize variables depending on xRange, yRange and selected_channels.
        """
        # xRange
        self._delta_with_buffer = self._scope.duration_buffer - self._xRange
        self._duration_plot_samples = math.ceil(
            self._xRange*self._scope.sample_rate)

    # ------------------------ Trigger Events ----------------------
    @abstractmethod
    def _update_LPT_trigger_events(self, trigger_arr):
        """
        Check if new LPT events (on the trigger channel) have entered the
        buffer. New events are added to self._trigger_events and displayed
        if needed.
        """
        pass

    def _clean_up_trigger_events(self):
        """
        Remove events exiting the buffer.
        """
        for k in range(len(self._trigger_events)-1, -1, -1):
            if self._trigger_events[k].position_buffer < 0:
                del self._trigger_events[k]

    # -------------------------- Main Loop -------------------------
    @abstractmethod
    def start_timer(self):
        """
        Start the update loop on a 20 ms timer.
        """
        pass

    @abstractmethod
    def _update_loop(self, *args, **kwargs):
        """
        Main update loop retrieving data from the scope's buffer and updating
        the Canvas.
        """
        self._scope.update_loop()

    # --------------------------- Events ---------------------------
    @abstractmethod
    def close(self):
        """
        Stops the update loop and close the window.
        """
        pass

    # --------------------------------------------------------------------
    @property
    def scope(self):
        """
        Scope connected to a StreamReceiver acquiring the data and applying
        filtering. The scope has a buffer of BUFFER_DURATION seconds
        (default: 30s).
        """
        return self._scope

    @property
    def xRange(self):
        """
        X-axis range/scale, i.e. the duration of the plotting window.
        """
        return self._xRange

    @xRange.setter
    @abstractmethod
    def xRange(self, xRange):
        """
        Called when the user changes the X-axis range/scale, i.e. the duration
        of the plotting window.
        """
        pass

    @property
    def yRange(self):
        """
        Y-axis range/scale, i.e. the signal amplitude.
        """
        return self._yRange

    @yRange.setter
    @abstractmethod
    def yRange(self, yRange):
        """
        Called when the user changes the signal range/scale.
        """
        pass

    @property
    def selected_channels(self):
        """
        Selected channels.
        """
        return self._selected_channels

    @selected_channels.setter
    @abstractmethod
    def selected_channels(self, selected_channels):
        """
        Called when the user changes the selection of channels.
        """
        pass

    @property
    def show_LPT_trigger_events(self):
        """
        Tick/Untick status of the show_LPT_trigger_events box.
        """
        return self._show_LPT_trigger_events

    @show_LPT_trigger_events.setter
    @abstractmethod
    def show_LPT_trigger_events(self, show_LPT_trigger_events):
        """
        Called when the user ticks or untick the show_LPT_trigger_events
        box.
        """
        pass


@fill_doc
class _Event(ABC):
    """
    Base class defining a trigger event.

    Parameters
    ----------
    %(viewer_event_type)s
    %(viewer_event_value)s
    %(viewer_position_buffer)s
    %(viewer_position_plot)s
    """
    _supported = ['LPT']

    @abstractmethod
    def __init__(self, event_type, event_value, position_buffer,
                 position_plot):
        assert event_type in self._supported
        self._event_type = event_type
        self._event_value = event_value
        self._position_buffer = position_buffer  # In time (s)
        self._position_plot = position_plot  # In time (s)

    @property
    def event_type(self):
        """
        Event type.
        """
        return self._event_type

    @property
    def event_value(self):
        """
        Event value.
        """
        return self._event_value

    @property
    def position_buffer(self):
        """
        Position in the buffer.
        """
        return self._position_buffer

    @position_buffer.setter
    def position_buffer(self, position_buffer):
        """
        Update both position in the buffer and the plotting window.
        """
        delta = self._position_buffer - position_buffer
        self._position_buffer = position_buffer
        self._position_plot -= delta

    @property
    def position_plot(self):
        """
        Position in the plotting window.
        """
        return self._position_plot

    @position_plot.setter
    def position_plot(self, position_plot):
        """
        Update only the position in the plotting window.
        """
        self._position_plot = position_plot



from configparser import ConfigParser

import pytest

from bsl import logger, set_log_level
from bsl.datasets import trigger_def, eeg_resting_state
from bsl.triggers import TriggerDef
from bsl.utils._tests import (requires_trigger_def_dataset,
                              requires_eeg_resting_state_dataset)


set_log_level('INFO')
logger.propagate = True


def _check_pair(tdef, name, value):
    """Check that the pair name, value is correctly set in tdef."""
    assert name in tdef._by_name
    assert value in tdef._by_value
    assert tdef._by_name[name] == value
    assert tdef._by_value[value] == name
    assert hasattr(tdef, name)
    assert getattr(tdef, name) == value


def test_trigger_def(caplog):
    """Test default behavior."""
    # Test with trigger_file = None
    tdef = TriggerDef()
    assert tdef._by_name == tdef.by_name == {}
    assert tdef._by_value == tdef.by_value == {}

    # Test add_event
    tdef.add('stim', 1, overwrite=False)
    _check_pair(tdef, 'stim', 1)

    # Test add existing event (name)
    caplog.clear()
    tdef.add('stim', 2, overwrite=False)
    _check_pair(tdef, 'stim', 1)
    assert 2 not in tdef._by_value
    assert 'Event name stim already exists. Skipping.' in caplog.text

    # Test add existing event (name - overwrite)
    tdef.add('stim', 2, overwrite=True)
    _check_pair(tdef, 'stim', 2)
    assert 1 not in tdef._by_value

    # Test add existing event (value)
    caplog.clear()
    tdef.add('rest', 2, overwrite=False)
    _check_pair(tdef, 'stim', 2)
    assert 'rest' not in tdef._by_name
    assert 'Event value 2 already exists. Skipping.' in caplog.text

    # Test add existing event (value - overwrite)
    tdef.add('rest', 2, overwrite=True)
    _check_pair(tdef, 'rest', 2)
    assert 'stim' not in tdef._by_name

    # Test remove non-existing event (name)
    caplog.clear()
    tdef.remove('stim')
    assert 'Event name stim not found.' in caplog.text

    # Test remove non-existing event (value)
    caplog.clear()
    tdef.remove(5)
    assert 'Event value 5 not found.' in caplog.text

    # Test remove existing event (name)
    tdef.remove('rest')
    assert 'rest' not in tdef._by_name
    assert 2 not in tdef._by_value
    assert not hasattr(tdef, 'rest')

    # Test remove existing event (value)
    tdef.add('rest', 2)
    tdef.remove(2)
    assert 'rest' not in tdef._by_name
    assert 2 not in tdef._by_value
    assert not hasattr(tdef, 'rest')


@requires_trigger_def_dataset
@requires_eeg_resting_state_dataset
def test_read_ini(caplog, tmp_path):
    """Test reading from a .ini file."""
    # Valid file
    tdef = TriggerDef(trigger_def.data_path())
    assert tdef._by_name == {'rest': 1,
                             'stim_left': 2,
                             'stim_right': 3,
                             'feedback': 4,
                             'start': 5,
                             'stop': 6}
    assert tdef._by_value == {1: 'rest',
                              2: 'stim_left',
                              3: 'stim_right',
                              4: 'feedback',
                              5: 'start',
                              6: 'stop'}

    # Valid .ini file with duplicate values
    caplog.clear()
    config = ConfigParser()
    config['events'] = {'stim': 1, 'rest': 1}
    with open(tmp_path / 'test.ini', 'w') as configfile:
        config.write(configfile)
    tdef = TriggerDef(tmp_path / 'test.ini')
    assert 'Event value %s already exists. Skipping.' % 1 in caplog.text
    assert tdef._by_name == {'stim': 1}
    assert tdef._by_value == {1: 'stim'}

    # Invalid file
    with pytest.raises(TypeError, match="'trigger_file' must be an instance"):
        tdef = TriggerDef(101)

    caplog.clear()
    tdef = TriggerDef(eeg_resting_state.data_path())
    assert tdef._trigger_file is None
    assert ('Argument trigger_file must be a valid Path to a .ini file. '
            'Provided: %s' % '.fif') in caplog.text

    caplog.clear()
    tdef = TriggerDef('non-existing-path')
    assert tdef._trigger_file is None
    assert ("Trigger event definition file '%s' not found."
            % 'non-existing-path') in caplog.text


def test_write_ini(tmp_path):
    """Test write to a .ini file."""
    # Valid file
    tdef = TriggerDef()
    tdef.add('stim', 1, overwrite=False)
    tdef.add('rest', 2, overwrite=False)
    tdef.write(tmp_path / 'test_write.ini')

    # Invalid file
    trigger_file = 101
    with pytest.raises(TypeError, match="'trigger_file' must be an instance"):
        tdef.write(trigger_file)

    trigger_file = tmp_path / 'test_write.txt'
    with pytest.raises(ValueError,
                       match='Argument trigger_file must end with .ini. '
                             'Provided: %s' % trigger_file.suffix):
        tdef.write(trigger_file)


@requires_trigger_def_dataset
def test_properties():
    """Test the properties."""
    tdef = TriggerDef(trigger_def.data_path())
    assert tdef._by_name == tdef.by_name == {'rest': 1,
                                             'stim_left': 2,
                                             'stim_right': 3,
                                             'feedback': 4,
                                             'start': 5,
                                             'stop': 6}
    assert tdef._by_value == tdef.by_value == {1: 'rest',
                                               2: 'stim_left',
                                               3: 'stim_right',
                                               4: 'feedback',
                                               5: 'start',
                                               6: 'stop'}

    with pytest.raises(AttributeError, match="can't set attribute"):
        tdef.by_name = dict()
    with pytest.raises(AttributeError, match="can't set attribute"):
        tdef.by_value = dict()



import time

import mne
import pytest

from bsl import StreamPlayer, logger, set_log_level
from bsl.datasets import eeg_resting_state
from bsl.externals import pylsl
from bsl.stream_receiver._stream import StreamMarker, StreamEEG  # noqa: F401
from bsl.stream_receiver._buffer import Buffer
from bsl.utils._tests import requires_eeg_resting_state_dataset


set_log_level('INFO')
logger.propagate = True


def _check_stream_properties(stream, streamInfo, stream_name, serial, is_slave,
                             lsl_time_offset, blocking, blocking_time,
                             dataset):
    """Check all the _Stream properties from a given stream."""
    raw = mne.io.read_raw_fif(fname=dataset.data_path(),
                              preload=True)

    # test streamInfo
    assert stream.streamInfo == stream._streamInfo == streamInfo
    with pytest.raises(AttributeError, match="can't set attribute"):
        stream.streamInfo = '101'

    # test sample rate
    assert stream.sample_rate == stream._sample_rate == raw.info['sfreq']
    with pytest.raises(AttributeError, match="can't set attribute"):
        stream.sample_rate = '2048'

    # test name
    assert stream.name == stream._name == stream_name
    with pytest.raises(AttributeError, match="can't set attribute"):
        stream.name = 'random-name'

    # test serial
    assert stream.serial == stream._serial == serial
    with pytest.raises(AttributeError, match="can't set attribute"):
        stream.serial = '101'

    # test is_slave
    assert stream.is_slave == stream._is_slave == bool(is_slave)
    with pytest.raises(AttributeError, match="can't set attribute"):
        stream.is_slave = not stream.is_slave

    # test channel list
    assert stream.ch_list == stream._ch_list == raw.ch_names
    with pytest.raises(AttributeError, match="can't set attribute"):
        stream.ch_list = [1, 0, 1]

    # test lsl_time_offset
    assert stream.lsl_time_offset == stream._lsl_time_offset == lsl_time_offset
    with pytest.raises(AttributeError, match="can't set attribute"):
        stream.lsl_time_offset = 101

    # test blocking
    assert stream.blocking == stream._blocking == bool(blocking)
    stream.blocking = not bool(blocking)
    assert stream.blocking == stream._blocking == (not bool(blocking))
    stream.blocking = bool(blocking)
    assert stream.blocking == stream._blocking == bool(blocking)

    # test blocking time
    assert stream.blocking_time == stream._blocking_time == blocking_time
    stream.blocking_time = blocking_time + 1
    assert stream.blocking_time == stream._blocking_time == blocking_time + 1
    stream.blocking_time = blocking_time
    assert stream.blocking_time == stream._blocking_time == blocking_time

    # test buffer
    assert isinstance(stream.buffer, Buffer)
    with pytest.raises(AttributeError, match="can't set attribute"):
        stream.buffer = 101


def test_stream_marker():
    """Test StreamMarker class used by the StreamReceiver."""
    # TODO
    pass


@requires_eeg_resting_state_dataset
def test_stream_eeg(caplog):
    """Test StreamEEG class used by the StreamReceiver."""
    # Default
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        streamInfos = pylsl.resolve_streams()
        for streamInfo in streamInfos:
            if streamInfo.name() == 'StreamPlayer':
                break

        stream = StreamEEG(streamInfo, bufsize=1, winsize=0.5)
        _check_stream_properties(stream=stream,
                                 streamInfo=streamInfo,
                                 stream_name='StreamPlayer',
                                 serial='N/A',
                                 is_slave=False,
                                 lsl_time_offset=None,
                                 blocking=True,
                                 blocking_time=5,
                                 dataset=eeg_resting_state)

        # test scaling factor
        assert stream.scaling_factor == stream._scaling_factor == 1
        stream.scaling_factor = 1e-6
        assert stream.scaling_factor == stream._scaling_factor == 1e-6
        stream.scaling_factor = 1
        assert stream.scaling_factor == stream._scaling_factor == 1
        with pytest.raises(ValueError,
                           match='Property scaling_factor must be a strictly '
                                 'positive number. Provided: 0'):
            stream.scaling_factor = 0
        with pytest.raises(ValueError,
                           match='Property scaling_factor must be a strictly '
                                 'positive number. Provided: -1'):
            stream.scaling_factor = -1

        # test acquire and bufsize/winsize
        time.sleep(1.1)
        stream.acquire()
        assert stream.lsl_time_offset is not None
        assert stream.buffer.bufsize == round(1 * stream.sample_rate) == 512
        assert stream.buffer.winsize == round(0.5 * stream.sample_rate) == 256
        assert len(stream.buffer.data) == len(stream.buffer.timestamps)
        assert len(stream.buffer.data) == stream.buffer.bufsize

    # Test disconnection/reconnection
    sp = StreamPlayer('StreamPlayer', eeg_resting_state.data_path())
    sp.start()

    streamInfos = pylsl.resolve_streams()
    for streamInfo in streamInfos:
        if streamInfo.name() == 'StreamPlayer':
            break

    caplog.clear()
    stream = StreamEEG(streamInfo, bufsize=1, winsize=0.5)
    stream.blocking = False
    stream.acquire()
    sp.stop()
    stream.acquire()
    assert 'Timeout occurred' not in caplog.text

    # Test disconnection timeout
    caplog.clear()
    sp.start()
    stream.blocking = True
    stream.blocking_time = 1
    sp.stop()
    stream.acquire()
    assert 'Timeout occurred [1.0secs] while acquiring data' in caplog.text



"""
Visualize LSL stream on the network. The signal is visualizd in real time with
spectral filtering and common average filtering option.

Command-line arguments:
    -s --stream_name    Stream name (str)

Example:
    bsl_stream_viewer
    bsl_stream_viewer -s StreamPlayer
"""

import argparse

from bsl import StreamViewer


def run():
    """Entrypoint for bsl_stream_viewer usage."""
    parser = argparse.ArgumentParser(
        prog='BSL StreamViewer',
        description='Starts a real-time viewer for a stream on LSL network.')
    parser.add_argument(
        '-s', '--stream_name', type=str, metavar='str',
        help='stream to display/plot.')

    args = parser.parse_args()
    stream_name = args.stream_name

    stream_viewer = StreamViewer(stream_name)
    stream_viewer.start()



import time
import pickle
from pathlib import Path

import mne
import pytest

from bsl import StreamRecorder, StreamPlayer, logger, set_log_level
from bsl.datasets import eeg_resting_state
from bsl.utils._tests import requires_eeg_resting_state_dataset


set_log_level('INFO')
logger.propagate = True


def _check_recorded_files(record_dir, eve_file, stream, fif_subdir):
    """Use eve_file to retrieve the file name stem and check if the recorded
    files exists."""
    fname_stem = eve_file.stem.split('-eve')[0]
    fname_pcl = record_dir / f'{fname_stem}-{stream}-raw.pcl'
    if fif_subdir:
        fname_fif = record_dir / 'fif' / f'{fname_stem}-{stream}-raw.fif'
    else:
        fname_fif = record_dir / f'{fname_stem}-{stream}-raw.fif'
    assert fname_pcl.exists()
    assert fname_fif.exists()


def _check_recorded_files_content(record_dir,  eve_file, stream, fif_subdir,
                                  dataset, record_duration):
    """Checks the recorded files content."""
    fname_stem = eve_file.stem.split('-eve')[0]
    fname_pcl = record_dir / f'{fname_stem}-{stream}-raw.pcl'
    if fif_subdir:
        fname_fif = record_dir / 'fif' / f'{fname_stem}-{stream}-raw.fif'
    else:
        fname_fif = record_dir / f'{fname_stem}-{stream}-raw.fif'
    raw = mne.io.read_raw_fif(fname=dataset.data_path(), preload=True)
    with open(fname_pcl, 'rb') as inp:
        raw_pcl = pickle.load(inp)
    raw_fif = mne.io.read_raw_fif(fname_fif, preload=True)
    assert raw.ch_names == raw_pcl['ch_names'] == raw_fif.ch_names
    assert raw.info['sfreq'] == raw_pcl['sample_rate'] == raw_fif.info['sfreq']
    assert raw_pcl['signals'].shape[::-1] == raw_fif.get_data().shape
    # some delay is introduce by the process operations
    assert 0 <= raw_fif.n_times / raw_fif.info['sfreq'] - record_duration < 0.2


@requires_eeg_resting_state_dataset
def test_stream_recorder(tmp_path, caplog):
    """Test recording capability of the stream recorder."""
    stream = 'StreamPlayer'
    record_duration = 0.5  # seconds
    dataset = eeg_resting_state
    fif_subdir = False

    # Test default call
    with StreamPlayer(stream, dataset.data_path()):
        recorder = StreamRecorder(record_dir=tmp_path,
                                  fif_subdir=fif_subdir,
                                  verbose=False)
        assert recorder._state.value == 0
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        assert eve_file is not None
        assert recorder._state.value == 1
        time.sleep(record_duration)
        recorder.stop()
        assert 'Waiting for StreamRecorder process to finish.' in caplog.text
        assert 'Recording finished.' in caplog.text
        assert recorder._eve_file is None
        assert recorder._process is None
        assert recorder._state.value == 0

    _check_recorded_files(tmp_path, eve_file, stream, fif_subdir)
    _check_recorded_files_content(tmp_path, eve_file, stream, fif_subdir,
                                  dataset, record_duration)

    # Test stop when not started
    caplog.clear()
    recorder = StreamRecorder(record_dir=tmp_path)
    assert recorder._state.value == 0
    recorder.stop()
    assert recorder._state.value == 0
    assert 'StreamRecorder was not started. Skipping.' in caplog.text

    # Test context manager
    with StreamPlayer(stream, dataset.data_path()):
        with StreamRecorder(record_dir=tmp_path, fname='test-context-manager',
                            fif_subdir=fif_subdir):
            time.sleep(record_duration)

    eve_file = tmp_path / 'test-context-manager-eve.txt'
    _check_recorded_files(tmp_path, eve_file, stream, fif_subdir)
    _check_recorded_files_content(tmp_path, eve_file, stream, fif_subdir,
                                  dataset, record_duration)


@requires_eeg_resting_state_dataset
def test_recording_multiple_streams(tmp_path):
    """Test multi-stream recording capabilities of the stream recorder."""
    record_duration = 0.5  # seconds
    dataset = eeg_resting_state
    fif_subdir = False

    with StreamPlayer('StreamPlayer1', dataset.data_path()), \
         StreamPlayer('StreamPlayer2', dataset.data_path()):

        # Record only StreamPlayer1
        recorder = StreamRecorder(record_dir=tmp_path,
                                  stream_name='StreamPlayer1',
                                  fif_subdir=fif_subdir,
                                  verbose=False)
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        time.sleep(record_duration)
        recorder.stop()

        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer1', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer1',
                                      fif_subdir, dataset, record_duration)

        # Record only StreamPlayer2
        recorder = StreamRecorder(record_dir=tmp_path,
                                  stream_name='StreamPlayer2',
                                  fif_subdir=fif_subdir,
                                  verbose=False)
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        time.sleep(record_duration)
        recorder.stop()

        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer2', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer2',
                                      fif_subdir, dataset, record_duration)

        # Record both - stream_name = None
        recorder = StreamRecorder(record_dir=tmp_path,
                                  stream_name=None,
                                  fif_subdir=fif_subdir,
                                  verbose=False)
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        time.sleep(record_duration)
        recorder.stop()

        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer1', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer1',
                                      fif_subdir, dataset, record_duration)
        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer2', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer2',
                                      fif_subdir, dataset, record_duration)

        # Record both - stream_name = ['StreamPlayer1', 'StreamPlayer2']
        recorder = StreamRecorder(record_dir=tmp_path,
                                  stream_name=['StreamPlayer1',
                                               'StreamPlayer2'],
                                  fif_subdir=fif_subdir,
                                  verbose=False)
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        time.sleep(record_duration)
        recorder.stop()

        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer1', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer1',
                                      fif_subdir, dataset, record_duration)
        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer2', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer2',
                                      fif_subdir, dataset, record_duration)


@requires_eeg_resting_state_dataset
def test_arg_fif_subdir(tmp_path):
    """Test argument fif_subdir."""
    record_duration = 0.5  # seconds
    dataset = eeg_resting_state

    with StreamPlayer('StreamPlayer', dataset.data_path()):

        # False
        fif_subdir = False

        recorder = StreamRecorder(record_dir=tmp_path, fif_subdir=fif_subdir,
                                  verbose=False)
        assert recorder._fif_subdir is False
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        time.sleep(record_duration)
        recorder.stop()
        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer',
                                      fif_subdir, dataset, record_duration)

        # True
        fif_subdir = True

        recorder = StreamRecorder(record_dir=tmp_path, fif_subdir=fif_subdir,
                                  verbose=False)
        assert recorder._fif_subdir is True
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        time.sleep(record_duration)
        recorder.stop()
        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer',
                                      fif_subdir, dataset, record_duration)


@requires_eeg_resting_state_dataset
def test_arg_verbose(tmp_path):
    """Test argument verbose."""
    record_duration = 0.5  # seconds
    dataset = eeg_resting_state
    fif_subdir = False

    with StreamPlayer('StreamPlayer', dataset.data_path()):

        # False
        verbose = False

        recorder = StreamRecorder(record_dir=tmp_path, fif_subdir=False,
                                  verbose=verbose)
        assert recorder._verbose is False
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        time.sleep(record_duration)
        recorder.stop()
        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer',
                                      fif_subdir, dataset, record_duration)

        # True
        verbose = True

        recorder = StreamRecorder(record_dir=tmp_path, fif_subdir=False,
                                  verbose=verbose)
        assert recorder._verbose is True
        recorder.start(blocking=True)
        eve_file = recorder.eve_file
        time.sleep(record_duration)
        recorder.stop()
        _check_recorded_files(tmp_path, eve_file, 'StreamPlayer', fif_subdir)
        _check_recorded_files_content(tmp_path, eve_file, 'StreamPlayer',
                                      fif_subdir, dataset, record_duration)


@requires_eeg_resting_state_dataset
def test_properties(tmp_path):
    """Test the StreamRecorder properties."""
    record_duration = 0.5  # seconds
    dataset = eeg_resting_state

    with StreamPlayer('StreamPlayer', dataset.data_path()):
        recorder = StreamRecorder(record_dir=tmp_path,
                                  fname='test',
                                  stream_name='StreamPlayer',
                                  fif_subdir=False,
                                  verbose=False)

        assert recorder.record_dir == tmp_path
        assert recorder.fname == 'test'
        assert recorder.stream_name == 'StreamPlayer'
        assert recorder.fif_subdir is False
        assert recorder.verbose is False

        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.record_dir = 'new path'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.fname = 'new fname'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.stream_name = 'new stream name'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.fif_subdir = True
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.verbose = True

        recorder.start(blocking=True)

        assert recorder.record_dir == tmp_path
        assert recorder.fname == 'test'
        assert recorder.stream_name == 'StreamPlayer'
        assert recorder.fif_subdir is False
        assert recorder.verbose is False

        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.record_dir = 'new path'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.fname = 'new fname'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.stream_name = 'new stream name'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.fif_subdir = True
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.verbose = True

        time.sleep(record_duration)
        recorder.stop()

        assert recorder.record_dir == tmp_path
        assert recorder.fname == 'test'
        assert recorder.stream_name == 'StreamPlayer'
        assert recorder.fif_subdir is False
        assert recorder.verbose is False

        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.record_dir = 'new path'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.fname = 'new fname'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.stream_name = 'new stream name'
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.fif_subdir = True
        with pytest.raises(AttributeError, match="can't set attribute"):
            recorder.verbose = True


def test_checker_arguments(tmp_path):
    """Test the argument error checking."""
    with pytest.raises(TypeError, match="'stream_name' must be an instance"):
        StreamRecorder(record_dir=tmp_path, stream_name=101)
    with pytest.raises(TypeError, match="'fif_subdir' must be an instance"):
        StreamRecorder(record_dir=tmp_path, fif_subdir=1)
    with pytest.raises(TypeError, match="'verbose' must be an instance"):
        StreamRecorder(record_dir=tmp_path, verbose=1)


def test_checker_record_dir(tmp_path):
    """Test the checker for argument record_dir."""
    # Valid
    recorder = StreamRecorder(record_dir=None)
    assert recorder.record_dir == Path.cwd()
    recorder = StreamRecorder(record_dir=tmp_path)
    assert recorder.record_dir == tmp_path

    # Invalid
    with pytest.raises(TypeError, match="'record_dir' must be an instance"):
        StreamRecorder(record_dir=101)


def test_checker_fname(tmp_path):
    """Test the checker for argument fname."""
    # Valid
    recorder = StreamRecorder(record_dir=tmp_path, fname=None)
    assert recorder.fname is None
    recorder = StreamRecorder(record_dir=tmp_path, fname='test')
    assert recorder.fname == 'test'

    # Invalid
    with pytest.raises(TypeError, match="'fname' must be an instance of"):
        StreamRecorder(record_dir=tmp_path, fname=101)


@requires_eeg_resting_state_dataset
def test_representation(tmp_path):
    """Test the representation method."""
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        recorder = StreamRecorder(record_dir=tmp_path, fname=None)
        expected = f'<Recorder: All streams | OFF | {tmp_path}>'
        assert recorder.__repr__() == expected
        recorder.start()
        expected = f'<Recorder: All streams | ON | {tmp_path}>'
        assert recorder.__repr__() == expected
        recorder.stop()
        expected = f'<Recorder: All streams | OFF | {tmp_path}>'
        assert recorder.__repr__() == expected

        recorder = StreamRecorder(record_dir=tmp_path, fname=None,
                                  stream_name='StreamPlayer')
        expected = f'<Recorder: StreamPlayer | OFF | {tmp_path}>'
        assert recorder.__repr__() == expected
        recorder.start()
        expected = f'<Recorder: StreamPlayer | ON | {tmp_path}>'
        assert recorder.__repr__() == expected
        recorder.stop()
        expected = f'<Recorder: StreamPlayer | OFF | {tmp_path}>'
        assert recorder.__repr__() == expected



"""
Trigger using software and a .txt file.
"""
from ._trigger import _Trigger
from ..externals import pylsl
from ..stream_recorder import StreamRecorder
from ..utils._checks import _check_type
from ..utils._docs import fill_doc, copy_doc


@fill_doc
class SoftwareTrigger(_Trigger):
    """
    Trigger saving signal value in a ``.txt`` file.
    Software trigger instance must be created after a StreamRecorder is started
    and close/deleted before a StreamRecorder is stopped.

    Parameters
    ----------
    recorder : StreamRecorder
        BSL's recorder used.
    %(trigger_verbose)s

    Examples
    --------
    >>> recorder = StreamRecorder('path to dir')
    >>> recorder.start()
    >>> trigger = SoftwareTrigger(recorder)
    >>> trigger.signal(1)
    >>> trigger.close() # OR >>> del trigger
    >>> recorder.stop()
    """

    def __init__(self, recorder, *, verbose: bool = True):
        super().__init__(verbose)
        self._recorder = SoftwareTrigger._check_recorder(recorder)
        self._eve_file = SoftwareTrigger._find_eve_file(recorder)
        try:
            self._eve_file = open(self._eve_file, 'a')
        except Exception:
            raise

    @copy_doc(_Trigger.signal)
    def signal(self, value: int) -> bool:
        _check_type(value, ('int', ), item_name='value')
        self._set_data(value)
        super().signal(value)
        return True

    @copy_doc(_Trigger._set_data)
    def _set_data(self, value: int):
        """
        Set the trigger signal to value.
        """
        super()._set_data(value)
        self._eve_file.write('%.6f\t0\t%d\n' % (pylsl.local_clock(), value))

    def close(self):
        """
        Close the event file.
        """
        try:
            self._eve_file.close()
        except Exception:
            pass

    def __del__(self):
        self.close()

    # --------------------------------------------------------------------
    @staticmethod
    def _check_recorder(recorder):
        """
        Check that the provided recorder is indeed a StreamRecorder.
        """
        _check_type(recorder, (StreamRecorder, ), item_name='recorder')
        return recorder

    @staticmethod
    def _find_eve_file(recorder):
        """
        Find the event file name from the on going StreamRecorder.
        """
        if recorder.eve_file is None:
            raise RuntimeError(
                'The StreamRecorder must be started before instantiating '
                'a SOFTWARE trigger.')

        return recorder.eve_file

    # --------------------------------------------------------------------
    @property
    def recorder(self):
        """
        BSL's recorder used.

        :type: StreamRecorder
        """
        return self._recorder

    @property
    def eve_file(self):
        """
        Event ``.ini`` file.

        :type: `~io.TextIOWrapper`
        """
        return self._eve_file



"""
Stream a recorded .fif file on LSL network.

Command-line arguments:
    #1                  Stream name (str)
    #2                  Raw FIF file to stream (str, path)
    -r --repeat         Number of repeat (int), default inf.
    -t --trigger_def    TriggerDef compatible file (str, path)
    -c --chunk_size     Chunk size (int)
    --high_resolution   Flag to use high resolution counter instead of sleep.

Examples:
    bsl_stream_player StreamPlayer "D:/Data/sample-raw.fif"
    bsl_stream_player StreamPlayer "D:/Data/sample-raw.fif" -r 5
    bsl_stream_player StreamPlayer "D:/Data/sample-raw.fif" -r 5 -c 16
    bsl_stream_player StreamPlayer "D:/Data/sample-raw.fif" --high_resolution
"""

import argparse

from bsl import StreamPlayer


def run():
    """Entrypoint for bsl_stream_player usage."""
    parser = argparse.ArgumentParser(
        prog='BSL StreamPlayer',
        description='Starts streaming data from a .fif on the network.')
    parser.add_argument(
        'stream_name', type=str,
        help='name of the stream displayed on LSL network.')
    parser.add_argument(
        'fif_file', type=str,
        help='path to the FIF File to stream on LSL network.')
    parser.add_argument(
        '-r', '--repeat', type=int, metavar='int',
        help='number of time the fif file is repeated.', default=float('inf'))
    parser.add_argument(
        '-t', '--trigger_def', type=str, metavar='str',
        help='path to a TriggerDef compatible file.', default=None)
    parser.add_argument(
        '-c', '--chunk_size', type=int, metavar='int',
        help='chunk size, usually 16 or 32.', default=16)
    parser.add_argument(
        '--high_resolution', action='store_true',
        help='use time.perf_counter() instead of time.sleep().')

    args = parser.parse_args()

    sp = StreamPlayer(args.stream_name, args.fif_file, args.repeat,
                      args.trigger_def, args.chunk_size, args.high_resolution)
    sp.start()
    input(">> Press ENTER to stop replaying data \n")
    sp.stop()



import math
import time

import mne
import pytest
import numpy as np

from bsl import StreamReceiver, StreamPlayer
from bsl.datasets import eeg_resting_state
from bsl.stream_viewer.scope import ScopeEEG
from bsl.stream_viewer.scope._scope import _BUFFER_DURATION
from bsl.utils._tests import requires_eeg_resting_state_dataset


@requires_eeg_resting_state_dataset
def test_scope_eeg():
    """Test EEG scope default capabilities."""
    bufsize = 0.2
    stream_name = 'StreamPlayer'

    with StreamPlayer(stream_name, eeg_resting_state.data_path()):
        receiver = StreamReceiver(bufsize=bufsize, winsize=bufsize,
                                  stream_name=stream_name)
        scope = ScopeEEG(receiver, stream_name)

        # test init of BP filter
        assert not hasattr(scope, '_sos')
        assert not hasattr(scope, '_zi_coeff')
        assert not hasattr(scope, '_zi')
        scope.init_bandpass_filter(low=1., high=40.)
        assert hasattr(scope, '_sos')
        assert hasattr(scope, '_zi_coeff')
        assert hasattr(scope, '_zi')
        assert scope._zi is None

        # test update loop
        assert len(scope._ts_list) == 0
        assert (scope._trigger_buffer == 0).all()
        assert (scope._data_buffer == 0).all()
        time.sleep(0.2)
        scope.update_loop()
        assert len(scope._ts_list) != 0
        assert (scope._data_buffer != 0).any()

        # car and BP were off
        assert scope._zi is None

        # test apply_car
        time.sleep(0.2)
        scope._apply_car = True
        scope.update_loop()

        # test apply BP
        time.sleep(0.2)
        scope._apply_bandpass = True
        scope.update_loop()
        assert scope._zi is not None

        # test change BP
        time.sleep(0.2)
        scope.init_bandpass_filter(low=1., high=20.)
        assert scope._zi is None
        scope.update_loop()
        assert scope._zi is not None

        # test selection of channel
        time.sleep(0.2)
        n = len(scope._selected_channels)
        scope._selected_channels = scope._selected_channels[:n//2]


@requires_eeg_resting_state_dataset
def test_buffer_duration():
    """Test the buffer size."""
    bufsize = 0.2
    stream_name = 'StreamPlayer'

    raw = mne.io.read_raw_fif(eeg_resting_state.data_path(), preload=False)
    sfreq = raw.info['sfreq']

    with StreamPlayer(stream_name, eeg_resting_state.data_path()):
        receiver = StreamReceiver(bufsize=bufsize, winsize=bufsize,
                                  stream_name=stream_name)
        scope = ScopeEEG(receiver, stream_name)
        assert scope.sample_rate == receiver.streams[stream_name].sample_rate \
            == sfreq

        assert scope.duration_buffer == _BUFFER_DURATION
        assert scope.duration_buffer_samples == math.ceil(
            _BUFFER_DURATION * sfreq)


@requires_eeg_resting_state_dataset
def test_properties():
    """Test EEG scope properties."""
    bufsize = 0.2
    stream_name = 'StreamPlayer'
    raw = mne.io.read_raw_fif(eeg_resting_state.data_path(), preload=False)

    with StreamPlayer(stream_name, eeg_resting_state.data_path()):
        receiver = StreamReceiver(bufsize=bufsize, winsize=bufsize,
                                  stream_name=stream_name)
        scope = ScopeEEG(receiver, stream_name)

        assert scope.stream_name == scope._stream_name == stream_name
        assert scope.sample_rate == scope._sample_rate
        assert scope.duration_buffer == scope._duration_buffer
        assert scope.duration_buffer_samples == scope._duration_buffer_samples
        assert scope.ts_list == scope._ts_list == list()
        assert scope.channels_labels == scope._channels_labels == \
            raw.ch_names[1:]
        assert scope.nb_channels == scope._nb_channels == len(raw.ch_names[1:])
        assert scope.apply_car == scope._apply_car
        assert not scope.apply_car
        assert scope.apply_bandpass == scope._apply_bandpass
        assert not scope.apply_bandpass
        assert scope.selected_channels == scope._selected_channels == \
            list(range(scope.nb_channels))
        assert (scope.data_buffer == scope._data_buffer).all()
        assert (scope.trigger_buffer == scope._trigger_buffer).all()

        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.stream_name = 'new name'
        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.sample_rate = 101
        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.duration_buffer = 101
        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.duration_buffer_samples = 101
        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.ts_list = [101]
        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.channels_labels = ['101']
        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.nb_channels = 101
        scope.apply_car = True
        assert scope.apply_car
        scope.apply_bandpass = True
        assert scope.apply_bandpass
        scope.selected_channels = list(range(scope.nb_channels // 2))
        assert scope.selected_channels == list(range(scope.nb_channels // 2))
        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.data_buffer = np.ones(
                (scope._nb_channels, scope._duration_buffer_samples),
                dtype=np.float32)
        with pytest.raises(AttributeError, match="can't set attribute"):
            scope.trigger_buffer = np.ones(scope._duration_buffer_samples)



import time

import mne

from bsl import StreamRecorder, StreamPlayer
from bsl.datasets import eeg_resting_state
from bsl.triggers import SoftwareTrigger
from bsl.utils._tests import requires_eeg_resting_state_dataset


@requires_eeg_resting_state_dataset
def test_trigger_software(tmp_path):
    """Test software triggers."""
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        recorder = StreamRecorder(record_dir=tmp_path, fname='test',
                                  stream_name='StreamPlayer', fif_subdir=False)
        recorder.start()

        trigger = SoftwareTrigger(recorder, verbose=True)
        assert trigger.verbose
        trigger.verbose = False
        time.sleep(0.5)

        assert trigger.signal(1)
        time.sleep(0.1)
        assert trigger.signal(2)

        time.sleep(0.5)
        trigger.close()
        time.sleep(0.5)
        recorder.stop()

    raw = mne.io.read_raw_fif(tmp_path / 'test-StreamPlayer-raw.fif')
    events = mne.find_events(raw, stim_channel='TRIGGER')
    assert events.shape == (2, 3)
    assert (events[:, 2] == [1, 2]).all()



import math
from abc import ABC, abstractmethod

from ...utils._logs import logger
from ...utils._docs import fill_doc

_BUFFER_DURATION = 30  # seconds


@fill_doc
class _Scope(ABC):
    """
    Class representing a base scope.

    Parameters
    ----------
    %(viewer_scope_stream_receiver)s
    %(viewer_scope_stream_name)s
    """

    # ---------------------------- INIT ----------------------------
    @abstractmethod
    def __init__(self, stream_receiver, stream_name):
        assert stream_name in stream_receiver.streams
        self._sr = stream_receiver
        self._stream_name = stream_name

        # Infos from stream
        self._sample_rate = int(
            self._sr.streams[self._stream_name].sample_rate)

        # Variables
        self._duration_buffer = _BUFFER_DURATION
        self._duration_buffer_samples = math.ceil(
            _BUFFER_DURATION*self._sample_rate)

        # Buffers
        self._ts_list = list()

        logger.debug('Scope connected to %s' % self._stream_name)
        logger.debug('Data sample rate is %f' % self._sample_rate)
        logger.debug(
            'Scope buffer duration is %d seconds' % self._duration_buffer)

    # -------------------------- Main Loop -------------------------
    @abstractmethod
    def update_loop(self):
        """
        Main update loop acquiring data from the LSL stream and filling the
        scope's buffer.
        """
        pass

    @abstractmethod
    def _read_lsl_stream(self):
        """
        Acquires data from the connected LSL stream.
        """
        self._sr.acquire()
        self._data_acquired, self._ts_list = self._sr._get_buffer()
        self._sr.reset_buffer()

        if len(self._ts_list) > 0:
            logger.debug('Signal acquired by the scope.')

    # --------------------------------------------------------------------
    @property
    def stream_name(self):
        """
        Name of the connected stream.
        """
        return self._stream_name

    @property
    def sample_rate(self):
        """
        Sample rate of the connected stream [Hz].
        """
        return self._sample_rate

    @property
    def duration_buffer(self):
        """
        Duration of the scope's buffer [seconds].
        """
        return self._duration_buffer

    @property
    def duration_buffer_samples(self):
        """
        Duration of the scope's buffer [samples].
        """
        return self._duration_buffer_samples

    @property
    def ts_list(self):
        """
        Timestamps buffer [samples, ].
        """
        return self._ts_list



"""
Module for signal visualization.

Visualize signals in real time with spectral filtering, common average
filtering options and real-time FFT.
"""

from .stream_viewer import StreamViewer  # noqa: F401



"""
PyQt5 Canvas for BSL's StreamViewer.
"""
import numpy as np

import pyqtgraph as pg
from PyQt5 import QtCore

from ._backend import _Backend, _Event
from ...utils._docs import fill_doc, copy_doc

# pg.setConfigOptions(antialias=True)


@fill_doc
class _BackendPyQtGraph(_Backend):
    """
    PyQtGraph backend for BSL's StreamViewer.

    Parameters
    ----------
    %(viewer_scope)s
    %(viewer_backend_geometry)s
    %(viewer_backend_xRange)s
    %(viewer_backend_yRange)s
    """

    # ---------------------------- Init ---------------------------
    def __init__(self, scope, geometry, xRange, yRange):
        super().__init__(scope, geometry, xRange, yRange)
        self._trigger_events = list()

        # Variables
        self._available_colors = np.random.uniform(
            size=(self._scope.nb_channels, 3), low=128, high=230)
        self._init_variables()

        # Canvas
        self._win = pg.GraphicsLayoutWidget(
            size=geometry[2:],
            title=f'Stream Viewer: {self._scope.stream_name}')
        self._win.show()
        self._plot_handler = self._win.addPlot()  # pyqtgraph.PlotItem
        subsampling_ratio = self._scope.sample_rate / 64
        self._plot_handler.setDownsampling(ds=subsampling_ratio,
                                           auto=None, mode='mean')
        self._plot_handler.setMouseEnabled(x=False, y=False)
        self._plot_handler.setMenuEnabled(False)
        self._init_canvas()

        # Plots
        self._plot_handler.clear()
        self._plots = dict()
        # Add PlotDataItem
        for k, idx in enumerate(self._scope.selected_channels):
            self._plots[idx] = self._plot_handler.plot(
                x=self._x_arr,
                y=self._scope.data_buffer[
                    idx, -self._duration_plot_samples:]+self._offset[k],
                pen=pg.mkColor(self._available_colors[idx, :]))

        # Timer
        self._timer = QtCore.QTimer(self._win)
        self._timer.timeout.connect(self._update_loop)

    @copy_doc(_Backend._init_variables)
    def _init_variables(self):
        super()._init_variables()

        # yRange
        self._offset = np.arange(
            0, -len(self._scope.selected_channels)*self._yRange,
            -self._yRange)

    def _init_canvas(self):
        """
        Initialize the drawing canvas.
        """
        # Ranges
        self._plot_handler.disableAutoRange()
        yRange = [
            1.5*self._yRange,
            -self._yRange*(len(self._scope.selected_channels)+1)]
        self._plot_handler.setRange(
            xRange=[0, self._xRange],
            yRange=yRange)
        self._plot_handler.showGrid(y=True)

        # Y-axis
        yticks = [(-k*self._yRange, self._scope.channels_labels[idx])
                  for k, idx in enumerate(self._scope.selected_channels)]
        ticks = [yticks, []]  # [major, minor]
        self._plot_handler.getAxis('left').setTicks(ticks)
        self._plot_handler.setLabel(
            axis='left', text=f'Scale (uV): {self._yRange}')

        # X-axis
        self._x_arr = np.arange(self._duration_plot_samples) \
            / self._scope.sample_rate
        self._plot_handler.setLabel(axis='bottom', text='Time (s)')

    # ------------------------ Trigger Events ----------------------
    @copy_doc(_Backend._update_LPT_trigger_events)
    def _update_LPT_trigger_events(self, trigger_arr):
        events_trigger_arr_idx = np.where(trigger_arr != 0)[0]
        events_values = trigger_arr[events_trigger_arr_idx]

        for k, event_value in enumerate(events_values):
            position_buffer = self._scope.duration_buffer - \
                (trigger_arr.shape[0] - events_trigger_arr_idx[k]) \
                / self._scope.sample_rate
            position_plot = position_buffer - self._delta_with_buffer

            event = _TriggerEvent(
                event_type='LPT',
                event_value=event_value,
                position_buffer=position_buffer,
                position_plot=position_plot,
                plot_handler=self._plot_handler,
                yRange=self._yRange)

            if position_plot >= 0:
                if event.event_type == 'LPT' and self._show_LPT_trigger_events:
                    event.addEventPlot()

            self._trigger_events.append(event)

    @copy_doc(_Backend._clean_up_trigger_events)
    def _clean_up_trigger_events(self):
        """
         Hide events exiting the plotting window.
        """
        super()._clean_up_trigger_events()
        for event in self._trigger_events:
            if event.position_plot < 0:
                event.removeEventPlot()

    # -------------------------- Main Loop -------------------------
    @copy_doc(_Backend.start_timer)
    def start_timer(self):
        self._timer.start(20)

    @copy_doc(_Backend._update_loop)
    def _update_loop(self):
        super()._update_loop()

        if len(self._scope.ts_list) > 0:
            for k, idx in enumerate(self._scope.selected_channels):
                self._plots[idx].setData(
                    x=self._x_arr,
                    y=self._scope.data_buffer[
                        idx, -self._duration_plot_samples:] + self._offset[k])

            # Update existing events position
            for event in self._trigger_events:
                event.position_buffer = event.position_buffer \
                    - len(self._scope.ts_list) / self._scope.sample_rate
            # Add new events entering the buffer
            self._update_LPT_trigger_events(
                self._scope.trigger_buffer[-len(self._scope.ts_list):])
            # Hide/Remove events exiting window and buffer
            self._clean_up_trigger_events()

    # --------------------------- Events ---------------------------
    @copy_doc(_Backend.close)
    def close(self):
        self._timer.stop()
        self._win.close()

    # ------------------------ Update program ----------------------
    @_Backend.xRange.setter
    @copy_doc(_Backend.xRange.setter)
    def xRange(self, xRange):
        self._xRange = xRange
        self._init_variables()
        self._init_canvas()

        for event in self._trigger_events:
            event.position_plot = event.position_buffer-self._delta_with_buffer
            if event.position_plot >= 0:
                if event.event_type == 'LPT' and self._show_LPT_trigger_events:
                    event.addEventPlot()
            else:
                event.removeEventPlot()

    @_Backend.yRange.setter
    @copy_doc(_Backend.yRange.setter)
    def yRange(self, yRange):
        self._yRange = yRange
        self._init_variables()
        self._init_canvas()

        for event in self._trigger_events:
            event.yRange = self._yRange

    @_Backend.selected_channels.setter
    @copy_doc(_Backend.selected_channels.setter)
    def selected_channels(self, selected_channels):
        plots2remove = [idx for idx in self._selected_channels
                        if idx not in selected_channels]
        plots2add = [idx for idx in selected_channels
                     if idx not in self._selected_channels]
        self._selected_channels = selected_channels
        self._init_variables()
        self._init_canvas()

        for idx in plots2remove:
            self._plot_handler.removeItem(self._plots[idx])
            del self._plots[idx]
        for k, idx in enumerate(plots2add):
            self._plots[idx] = self._plot_handler.plot(
                x=self._x_arr,
                y=self._scope.data_buffer[
                    idx, -self._duration_plot_samples:] + self._offset[k],
                pen=pg.mkColor(self._available_colors[idx, :]))

    @_Backend.show_LPT_trigger_events.setter
    @copy_doc(_Backend.show_LPT_trigger_events.setter)
    def show_LPT_trigger_events(self, show_LPT_trigger_events):
        self._show_LPT_trigger_events = show_LPT_trigger_events
        for event in self._trigger_events:
            if event.position_plot >= 0 and event.event_type == 'LPT':
                if self._show_LPT_trigger_events:
                    event.addEventPlot()
                else:
                    event.removeEventPlot()


@fill_doc
class _TriggerEvent(_Event):
    """
    Class defining a trigger event for the pyqtgraph backend.

    Parameters
    ----------
    %(viewer_event_type)s
    %(viewer_event_value)s
    %(viewer_position_buffer)s
    %(viewer_position_plot)s
    plot_handler : pyqtgraph.PlotItem
        Plot handler.
    yRange : int | float
        Currently set signal range/scale.
    """
    colors = {'LPT': pg.mkColor(0, 255, 0)}

    def __init__(self, event_type, event_value, position_buffer, position_plot,
                 plot_handler, yRange):
        super().__init__(event_type, event_value,
                         position_buffer, position_plot)
        self._plot_handler = plot_handler
        self._yRange = yRange

        self._lineItem = None
        self._textItem = None
        self._plotted = False

    def addEventPlot(self):
        """
        Plots the event on the handler.
        """
        if not self._plotted:
            self._lineItem = pg.InfiniteLine(
                pos=self._position_plot, pen=self.colors[self._event_type])
            self._plot_handler.addItem(self._lineItem)

            self._textItem = pg.TextItem(str(self._event_value),
                                         anchor=(0.5, 0.5),
                                         fill=(0, 0, 0),
                                         color=self.colors[self._event_type])
            self._textItem.setPos(self._position_plot, 1.5*self._yRange)
            self._plot_handler.addItem(self._textItem)
            self._plotted = True

    def removeEventPlot(self):
        """
        Remove the event from the plot handler.
        """
        if self._plotted:
            self._plot_handler.removeItem(self._lineItem)
            self._plot_handler.removeItem(self._textItem)
            self._lineItem = None
            self._textItem = None
            self._plotted = False

    def _update(self):
        """
        Updates the plot handler.
        """
        if self._lineItem is not None:
            self._lineItem.setValue(self._position_plot)
        if self._textItem is not None:
            self._textItem.setPos(self._position_plot, 1.5*self._yRange)

    def __del__(self):
        try:
            self.removeEventPlot()
        except Exception:
            pass

    @property
    @copy_doc(_Event.event_type)
    def event_type(self):
        return self._event_type

    @property
    @copy_doc(_Event.event_value)
    def event_value(self):
        return self._event_value

    @_Event.position_buffer.setter
    @copy_doc(_Event.position_buffer.setter)
    def position_buffer(self, position_buffer):
        _Event.position_buffer.__set__(self, position_buffer)
        self._update()

    @_Event.position_plot.setter
    @copy_doc(_Event.position_plot.setter)
    def position_plot(self, position_plot):
        _Event.position_plot.__set__(self, position_plot)
        self._update()

    @property
    def plotted(self):
        """
        True if the event is displayed, else False.
        """
        return self._plotted

    @property
    def yRange(self):
        """
        Signal range/scale used to position the TextItem.
        """
        return self._yRange

    @yRange.setter
    def yRange(self, yRange):
        """
        Impacts the position of the TextItem.
        """
        self._yRange = yRange
        self._update()



from ..utils._logs import logger
from ..utils._docs import fill_doc


class Buffer():
    """
    Class representing the stream's buffer.
    This is not a Python Buffer Interface compatible class.

    Parameters
    ----------
    bufsize : int
        Buffer's size [samples].
    winsize : int
        Window's size [samples]. Must be smaller than the buffer's size.
    """

    def __init__(self, bufsize, winsize):
        self._bufsize = bufsize
        self._winsize = winsize

        self._data = []
        self._timestamps = []

    @fill_doc
    def fill(self, data, tslist):
        """
        Fill the data and timestamps to the buffer.

        Parameters
        -----------
        data : list
            Received data [samples x channels].
        %(receiver_tslist)s
        """
        self._data.extend(data)
        self._timestamps.extend(tslist)
        logger.debug('Buffer filled with %d points.' % len(tslist))

        if len(self._timestamps) > self._bufsize:
            self._data = self._data[-self._bufsize:]
            self._timestamps = self._timestamps[-self._bufsize:]

    def reset_buffer(self):
        """
        Clear the buffer's data and timestamps.
        """
        self._data = []
        self._timestamps = []
        logger.debug('Buffer reset.')

    @property
    def bufsize(self):
        """
        Buffer's size [samples].
        """
        return self._bufsize

    @property
    def winsize(self):
        """
        Window's size [samples].
        """
        return self._winsize

    @property
    def data(self):
        """
        Buffer's data [samples x channels].
        """
        return self._data

    @property
    def timestamps(self):
        """
        Data's timestamps [samples].
        """
        return self._timestamps



from bsl.stream_receiver._buffer import Buffer


def test_buffer():
    """Test Buffer class used by the StreamReceiver."""
    # buffer creation
    b = Buffer(bufsize=100, winsize=80)
    assert b.bufsize == b._bufsize == 100
    assert b.winsize == b._winsize == 80
    assert len(b.data) == len(b._data) == 0
    assert len(b.timestamps) == len(b._timestamps) == 0

    # add elements less than buffer limit
    chs = 3
    samples = 5
    fake_data = [[1] * chs] * samples
    fake_timestamps = list(range(samples))
    b.fill(fake_data, fake_timestamps)
    assert len(b.data) == len(b._data) == samples
    assert len(b.timestamps) == len(b._timestamps) == samples
    assert all(elt == [1]*chs for elt in b.data)
    assert b.timestamps == fake_timestamps

    # add elements above buffer limit
    samples = 97
    fake_data = [[2] * chs] * samples
    fake_timestamps = list(range(samples))
    b.fill(fake_data, fake_timestamps)
    assert len(b.data) == len(b._data) == b.bufsize
    assert len(b.timestamps) == len(b._timestamps) == b.bufsize
    assert all(elt == [1]*chs for elt in b.data[:3])
    assert all(elt == [2]*chs for elt in b.data[3:])

    # reset buffer
    b.reset_buffer()
    assert len(b.data) == len(b._data) == 0
    assert len(b.timestamps) == len(b._timestamps) == 0



"""
Sample EEG dataset recorded on ANT Neuro Amplifier with 64 electrodes. The
recording last 184 seconds and include 75 rest events (1)  lasting 1 second and
75 auditory stimuli events (4) lasting 0.8 second.
"""

import os
from pathlib import Path

from ._fetching import fetch_file, _hashfunc
from ..utils._logs import logger


MD5 = 'ea0d40643bdc1c88e2b808c7128a0eba'
SHA1 = 'd86e0ad1d53224a41e06b918e5d709615e2da32a'
URL = 'https://github.com/bsl-tools/bsl-datasets/raw/main/eeg_sample/auditory_stimuli-raw.fif'  # noqa: E501
PATH = Path('~/bsl_data/eeg_sample/auditory_stimuli-raw.fif').expanduser()


def data_path():
    """
    Path to a sample EEG dataset recorded on ANT Neuro Amplifier with 64
    electrodes. The recording last 184 seconds and include 75 rest events (1)
    lasting 1 second and 75 auditory stimuli events (4) lasting 0.8 second.
    If the dataset is not locally present, it is downloaded in the user home
    directory in the folder ``bsl_data/eeg_sample``.

    Returns
    -------
    path : Path
        Path to the dataset.
    """
    os.makedirs(PATH.parent, exist_ok=True)

    logger.debug('URL:   %s' % (URL,))
    logger.debug('Hash:  %s' % (MD5,))
    logger.debug('Path:  %s' % (PATH,))

    if PATH.exists() and _hashfunc(PATH, hash_type='md5') == MD5:
        download = False
    elif PATH.exists() and not _hashfunc(PATH, hash_type='md5') == MD5:
        logger.warning(
            'Dataset existing but with different hash. Re-downloading.')
        download = True
    else:
        logger.info('Fetching dataset..')
        download = True

    if download:
        fetch_file(URL, PATH, hash_=MD5, hash_type='md5', timeout=10.)

    return PATH



"""
Module for signal replay. Replay the recorded signals in real time as if it was
transmitted from a real acquisition server.

For Windows users, make sure to use the provided time resolution
tweak tool (TimerTool) to set to 500us the time resolution of the OS.
"""

from .stream_player import StreamPlayer  # noqa: F401


__all__ = ['StreamPlayer']



import re
import time

import mne
import pytest
import numpy as np

from bsl import StreamReceiver, StreamPlayer, logger, set_log_level
from bsl.datasets import eeg_resting_state
from bsl.utils._tests import requires_eeg_resting_state_dataset


set_log_level('INFO')
logger.propagate = True


@requires_eeg_resting_state_dataset
def test_stream_receiver():
    """Test stream receiver default functionalities."""
    stream = 'StreamPlayer'

    with StreamPlayer(stream, eeg_resting_state.data_path()):
        sr = StreamReceiver(bufsize=1, winsize=0.2, stream_name=stream)
        assert stream in sr.streams
        assert sr._connected

        # test infos
        info = sr.mne_infos[stream]
        assert isinstance(info, mne.io.Info)
        assert info.ch_names == sr.streams[stream].ch_list
        assert info['sfreq'] == sr.streams[stream].sample_rate
        info._check_consistency()

        # test get_xxx numpy
        time.sleep(1)
        sr.acquire()
        window, timestamps_window = sr.get_window()
        buffer, timestamps_buffer = sr.get_buffer()
        assert window.shape[1] == buffer.shape[1] == \
            len(sr.streams[stream].ch_list)
        assert window.shape[0] == len(timestamps_window) == \
            round(sr.streams[stream].sample_rate * 0.2)
        assert buffer.shape[0] == len(timestamps_buffer) == \
            round(sr.streams[stream].sample_rate * 1)

        window2, timestamps_window2 = sr.get_window()
        assert np.all(window == window2)
        assert np.all(timestamps_window == timestamps_window2)

        time.sleep(0.2)
        previous_buffer = buffer
        sr.acquire()
        window3, _ = sr.get_window()
        buffer3, _ = sr.get_buffer()
        assert window3[0, :] in previous_buffer

        # test get_window MNE raw
        raw, _ = sr.get_window(return_raw=True)
        assert isinstance(raw, mne.io.BaseRaw)
        assert raw.info.ch_names == info.ch_names
        assert raw.info['sfreq'] == info['sfreq']
        raw.info._check_consistency()
        data = raw.get_data(picks='all').T
        assert np.all(data == window3)

        # test get_buffer MNE raw
        raw, _ = sr.get_buffer(return_raw=True)
        assert isinstance(raw, mne.io.BaseRaw)
        assert raw.info.ch_names == info.ch_names
        assert raw.info['sfreq'] == info['sfreq']
        raw.info._check_consistency()
        data = raw.get_data(picks='all').T
        assert np.all(data == buffer3)

        del sr


@requires_eeg_resting_state_dataset
def test_receiving_multi_streams():
    """Test StreamReceiver multi-streams functionalities."""
    with StreamPlayer('StreamPlayer1', eeg_resting_state.data_path()), \
         StreamPlayer('StreamPlayer2', eeg_resting_state.data_path()):

        # test connect to only one
        sr = StreamReceiver(bufsize=1, winsize=0.2,
                            stream_name='StreamPlayer1')
        assert 'StreamPlayer1' in sr.streams
        assert sr.stream_name == ['StreamPlayer1']
        del sr

        # test connect to both
        sr = StreamReceiver(bufsize=1, winsize=0.2)
        assert 'StreamPlayer1' in sr.streams
        assert 'StreamPlayer2' in sr.streams
        assert sr.stream_name is None

        # test get_xxx numpy
        time.sleep(1)
        sr.acquire()
        window1, timestamps_window1 = sr.get_window('StreamPlayer1')
        buffer1, timestamps_buffer1 = sr.get_buffer('StreamPlayer1')
        window2, timestamps_window2 = sr.get_window('StreamPlayer2')
        buffer2, timestamps_buffer2 = sr.get_buffer('StreamPlayer2')
        assert window1.shape[1] == window2.shape[1] == \
               buffer1.shape[1] == buffer2.shape[1] == \
               len(sr.streams['StreamPlayer1'].ch_list) == \
               len(sr.streams['StreamPlayer2'].ch_list)
        assert window1.shape[0] == window2.shape[0] == \
               len(timestamps_window1) == len(timestamps_window2) == \
               round(sr.streams['StreamPlayer1'].sample_rate * 0.2) == \
               round(sr.streams['StreamPlayer2'].sample_rate * 0.2)
        assert buffer1.shape[0] == buffer2.shape[0] == \
               len(timestamps_buffer1) == len(timestamps_buffer2) == \
               round(sr.streams['StreamPlayer1'].sample_rate * 1) == \
               round(sr.streams['StreamPlayer2'].sample_rate * 1)

        # test get_window MNE raw
        raw1, _ = sr.get_window('StreamPlayer1', return_raw=True)
        raw2, _ = sr.get_window('StreamPlayer2', return_raw=True)
        assert isinstance(raw1, mne.io.BaseRaw)
        assert isinstance(raw2, mne.io.BaseRaw)
        info1 = sr.mne_infos['StreamPlayer1']
        info2 = sr.mne_infos['StreamPlayer2']
        assert raw1.info.ch_names == raw2.info.ch_names == \
               info1.ch_names == info2.ch_names
        assert raw1.info['sfreq'] == raw2.info['sfreq'] == \
               info1['sfreq'] == info2['sfreq']
        raw1.info._check_consistency()
        raw2.info._check_consistency()
        data1 = raw1.get_data(picks='all').T
        data2 = raw2.get_data(picks='all').T
        assert np.all(data1 == window1)
        assert np.all(data2 == window2)

        # test get_buffer MNE raw
        raw1, _ = sr.get_buffer('StreamPlayer1', return_raw=True)
        raw2, _ = sr.get_buffer('StreamPlayer2', return_raw=True)
        assert isinstance(raw1, mne.io.BaseRaw)
        assert isinstance(raw2, mne.io.BaseRaw)
        info1 = sr.mne_infos['StreamPlayer1']
        info2 = sr.mne_infos['StreamPlayer2']
        assert raw1.info.ch_names == raw2.info.ch_names == \
               info1.ch_names == info2.ch_names
        assert raw1.info['sfreq'] == raw2.info['sfreq'] == \
               info1['sfreq'] == info2['sfreq']
        raw1.info._check_consistency()
        raw2.info._check_consistency()
        data1 = raw1.get_data(picks='all').T
        data2 = raw2.get_data(picks='all').T
        assert np.all(data1 == buffer1)
        assert np.all(data2 == buffer2)

        del sr


@requires_eeg_resting_state_dataset
def test_properties():
    """Test the StreamReceiver properties."""
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        sr = StreamReceiver(bufsize=1, winsize=0.2)
        assert sr.winsize == sr._winsize == 0.2
        assert sr.bufsize == sr._bufsize == 1
        assert sr.stream_name is None
        assert sr.connected == sr._connected
        assert sr.connected
        assert sr.mne_infos == sr._mne_infos
        assert isinstance(sr.mne_infos, dict)
        assert len(sr.mne_infos) == 1
        assert isinstance(sr.mne_infos['StreamPlayer'], mne.io.Info)
        assert sr.streams == sr._streams
        assert isinstance(sr.streams, dict)
        assert len(sr.streams) == 1
        assert 'StreamPlayer' in sr.streams

        with pytest.raises(AttributeError, match="can't set attribute"):
            sr.winsize = 1
        with pytest.raises(AttributeError, match="can't set attribute"):
            sr.bufsize = 2
        with pytest.raises(AttributeError, match="can't set attribute"):
            sr.stream_name = 'StreamPlayer'
        with pytest.raises(AttributeError, match="can't set attribute"):
            sr.connected = False
        with pytest.raises(AttributeError, match="can't set attribute"):
            sr.mne_infos = dict()
        with pytest.raises(AttributeError, match="can't set attribute"):
            sr.streams = dict()

        sr.disconnect()
        assert sr.streams == dict()
        assert sr.connected is False

        del sr


@requires_eeg_resting_state_dataset
def test_get_method_warning_and_errors(caplog):
    """Test the checking done in get_xxx methods."""
    with StreamPlayer('StreamPlayer1', eeg_resting_state.data_path()), \
         StreamPlayer('StreamPlayer2', eeg_resting_state.data_path()):
        sr = StreamReceiver(bufsize=1, winsize=1, stream_name=None)

        # Disconnect and acquire
        sr.disconnect()
        with pytest.raises(RuntimeError,
                           match='StreamReceiver is not connected to any '
                                 'streams.'):
            sr.acquire()

        # Acquire, disconnect and get.
        sr.connect()
        time.sleep(1)
        sr.acquire()
        time.sleep(0.05)
        sr.disconnect()
        with pytest.raises(RuntimeError,
                           match='StreamReceiver is not connected to any '
                           'streams.'):
            sr.get_window()
        with pytest.raises(RuntimeError,
                           match='StreamReceiver is not connected to any '
                                 'streams.'):
            sr.get_buffer()

        # multiple streams
        sr.connect()
        time.sleep(1)
        sr.acquire()
        with pytest.raises(RuntimeError,
                           match='StreamReceiver is connected to multiple '
                                 'streams. Please provide the stream_name '
                                 'argument.'):
            sr.get_window()
        with pytest.raises(RuntimeError,
                           match='StreamReceiver is connected to multiple '
                                 'streams. Please provide the stream_name '
                                 'argument.'):
            sr.get_buffer()

        # stream_name not in streams
        with pytest.raises(KeyError,
                           match="StreamReceiver is not connected to '101'."):
            sr.get_window(stream_name='101')
        with pytest.raises(KeyError,
                           match="StreamReceiver is not connected to '101'."):
            sr.get_buffer(stream_name='101')

        # forgot to call .acquire()
        sr.disconnect()
        sr.connect()
        with pytest.raises(AttributeError,
                           match=re.escape(
                               '.acquire() must be called before '
                               '.get_window().')):
            sr.get_window(stream_name='StreamPlayer1')
        with pytest.raises(AttributeError,
                           match=re.escape(
                               '.acquire() must be called before '
                               '.get_buffer().')):
            sr.get_buffer(stream_name='StreamPlayer1')

        # not enough samples
        del sr
        sr = StreamReceiver(bufsize=1, winsize=1, stream_name=None)
        # caplog.clear()
        sr.acquire()
        sr.get_window(stream_name='StreamPlayer1')
        assert ('The buffer of StreamPlayer1 does not contain enough samples. '
                'Returning the available samples.') in caplog.text

        # could not convert to MNE raw
        time.sleep(1)
        sr.acquire()
        caplog.clear()
        sr._mne_infos['StreamPlayer1'] = None
        data, _ = sr.get_window(stream_name='StreamPlayer1', return_raw=True)
        assert ('The LSL stream StreamPlayer1 can not be converted to MNE raw '
                'instance. Returning numpy arrays.') in caplog.text
        assert isinstance(data, np.ndarray)
        caplog.clear()
        data, _ = sr.get_buffer(stream_name='StreamPlayer1', return_raw=True)
        assert ('The LSL stream StreamPlayer1 can not be converted to MNE raw '
                'instance. Returning numpy arrays.') in caplog.text
        assert isinstance(data, np.ndarray)

        del sr


@requires_eeg_resting_state_dataset
def test_connect_disconnect():
    """Test connect and disconnect methods."""
    with StreamPlayer('StreamPlayer1', eeg_resting_state.data_path()), \
         StreamPlayer('StreamPlayer2', eeg_resting_state.data_path()):
        sr = StreamReceiver(bufsize=1, winsize=1, stream_name='StreamPlayer1')
        assert sr.connected
        assert sr.stream_name == ['StreamPlayer1']
        assert 'StreamPlayer1' in sr.streams
        assert 'StreamPlayer2' not in sr.streams

        sr.connect('StreamPlayer2', force=False)
        assert sr.connected
        assert sr.stream_name == ['StreamPlayer1']
        assert 'StreamPlayer1' in sr.streams
        assert 'StreamPlayer2' not in sr.streams

        sr.connect('StreamPlayer2', force=True)
        assert sr.connected
        assert sr.stream_name == ['StreamPlayer1', 'StreamPlayer2']
        assert 'StreamPlayer1' in sr.streams
        assert 'StreamPlayer2' in sr.streams

        sr.disconnect('StreamPlayer1')
        assert sr.connected
        assert sr.stream_name == ['StreamPlayer2']
        assert 'StreamPlayer1' not in sr.streams
        assert 'StreamPlayer2' in sr.streams

        sr.connect('StreamPlayer1', force=True)
        assert sr.connected
        assert sr.stream_name == ['StreamPlayer2', 'StreamPlayer1']
        assert 'StreamPlayer1' in sr.streams
        assert 'StreamPlayer2' in sr.streams

        sr.disconnect()
        assert not sr.connected
        assert sr.stream_name == list()
        assert 'StreamPlayer1' not in sr.streams
        assert 'StreamPlayer2' not in sr.streams

        sr.connect(stream_name=None, force=True)
        assert sr.connected
        assert sr.stream_name is None
        assert 'StreamPlayer1' in sr.streams
        assert 'StreamPlayer2' in sr.streams

        del sr


@requires_eeg_resting_state_dataset
def test_checker_bufsize(caplog):
    """Test the checker for argument bufsize."""
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        # Valid
        sr = StreamReceiver(bufsize=1, winsize=0.2)
        assert isinstance(sr.bufsize, float)
        assert sr.bufsize == 1.0
        del sr

        # Invalid value
        with pytest.raises(ValueError,
                           match='Argument bufsize must be a strictly'
                                 ' positive int or a float. '
                                 'Provided: %s' % -101):
            StreamReceiver(bufsize=-101, winsize=0.2)

        # Invalid type
        with pytest.raises(TypeError, match="'bufsize' must be an instance"):
            StreamReceiver(bufsize=[101], winsize=0.2)

        # Smaller than winsize
        caplog.clear()
        sr = StreamReceiver(bufsize=1, winsize=2)
        assert isinstance(sr.bufsize, float)
        assert isinstance(sr.winsize, float)
        assert sr.bufsize == 2.0
        assert sr.winsize == 2.0
        assert ('Buffer size %.2f is smaller than window size. ' % 1.0 +
                'Setting to %.2f.' % 2.0) in caplog.text

        del sr


@requires_eeg_resting_state_dataset
def test_checker_winsize():
    """Test the checker for argument winsize."""
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        # Valid
        sr = StreamReceiver(bufsize=1, winsize=0.2)
        assert isinstance(sr.winsize, float)
        assert sr.winsize == 0.2
        del sr

        # Invalid value
        with pytest.raises(ValueError,
                           match='Argument winsize must be a strictly positive'
                                 ' int or a float. Provided: %s' % -101):
            StreamReceiver(bufsize=1, winsize=-101)

        # Invalid type
        with pytest.raises(TypeError, match="'winsize' must be an instance"):
            StreamReceiver(bufsize=1, winsize=[101])


@requires_eeg_resting_state_dataset
def test_checker_stream_name(caplog):
    """Test the checker for argument stream_name."""
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        # Valid - None
        sr = StreamReceiver(bufsize=1, winsize=0.2, stream_name=None)
        assert sr.stream_name is None
        del sr

        # Valid - str
        sr = StreamReceiver(bufsize=1, winsize=0.2, stream_name='StreamPlayer')
        assert sr.stream_name == ['StreamPlayer']
        del sr

        # Valid - list of str
        sr = StreamReceiver(bufsize=1, winsize=0.2,
                            stream_name=['StreamPlayer'])
        assert sr.stream_name == ['StreamPlayer']
        del sr

        # Valid - tuple of str
        sr = StreamReceiver(bufsize=1, winsize=0.2,
                            stream_name=('StreamPlayer', ))
        assert sr.stream_name == ['StreamPlayer']
        del sr

        # Invalid - list of 'not str'
        with pytest.raises(TypeError,
                           match=re.escape(
                               'Argument stream_name must be a string or a '
                               'list of strings. Provided: [1, 0, 1]')):
            StreamReceiver(bufsize=1, winsize=0.2, stream_name=[1, 0, 1])

        # Invalid type
        with pytest.raises(TypeError,
                           match="'stream_name' must be an instance"):
            StreamReceiver(bufsize=1, winsize=0.2, stream_name=101)


@requires_eeg_resting_state_dataset
def test_representation():
    """Test the representation method."""
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        sr = StreamReceiver(bufsize=1, winsize=0.2)
        expected = "<Receiver: ('StreamPlayer',) | ON | buf: 1.0s - win: 0.2s>"
        assert sr.__repr__() == expected

        sr.disconnect()
        expected = "<Receiver: () | OFF | buf: 1.0s - win: 0.2s>"
        assert sr.__repr__() == expected

        # Explicit stream name.
        sr.connect(stream_name='StreamPlayer')
        expected = "<Receiver: ('StreamPlayer',) | ON | buf: 1.0s - win: 0.2s>"
        assert sr.__repr__() == expected

        sr.disconnect()
        expected = "<Receiver: () | OFF | buf: 1.0s - win: 0.2s>"
        assert sr.__repr__() == expected

        del sr



"""
External packages bundled together with BSL.
"""
import importlib


try:
    pylsl = importlib.import_module(name='pylsl')
except (ModuleNotFoundError, ImportError, RuntimeError):
    pylsl = importlib.import_module(name='.pylsl', package=__name__)



import pytest

from bsl import logger, set_log_level
from bsl.triggers import MockTrigger


set_log_level('INFO')
logger.propagate = True


def test_trigger_mock(caplog):
    """Testing for Mock triggers."""
    trigger = MockTrigger(verbose=True)
    assert trigger.verbose
    trigger.verbose = False
    assert trigger.signal(1)
    assert 'MOCK trigger set to 1' in caplog.text
    assert trigger.signal(2)
    assert 'MOCK trigger set to 2' in caplog.text

    with pytest.raises(TypeError, match="'value' must be an instance"):
        trigger.signal(3.0)
    assert 'MOCK trigger set to 3.0' not in caplog.text



"""
LSL wrapper functions for creating a server and a client.
"""
import xml.etree.ElementTree as ET

from . import Timer
from ._checks import _check_type
from ._logs import logger
from ..externals import pylsl


def list_lsl_streams(ignore_markers=False):
    """
    List all the available outlets on LSL network.

    Parameters
    ----------
    ignore_markers : bool
        If ``True``, ignore streams with Marker type.

    Returns
    -------
    stream_list : list
        List of the found stream name.
    streamInfos : list
        List of the corresponding ``pylsl.StreamInfo``.
    """
    _check_type(ignore_markers, (bool, ), item_name='ignore_markers')

    stream_list = []
    streamInfos = pylsl.resolve_streams()

    if len(streamInfos) == 0:
        return stream_list, []

    for index, streamInfo in enumerate(streamInfos):
        if ignore_markers and 'Markers' in streamInfo.type():
            continue
        stream_list.append(streamInfo.name())

    if ignore_markers:
        streamInfos = [streamInfo for streamInfo in streamInfos
                       if 'Markers' not in streamInfo.type()]

    return stream_list, streamInfos


def search_lsl(ignore_markers=False, timeout=10):
    """
    Search and select an available stream on LSL network.
    Does not open an LSL inlet.

    Parameters
    ----------
    ignore_markers : bool
        If ``True``, ignore streams with Marker type.
    timeout : int
        Timeout duration in seconds after which the search of an LSL stream
        is interrupted.

    Returns
    -------
    stream_name : str
        Selected stream name.
    """
    _check_type(ignore_markers, (bool, ), item_name='ignore_markers')
    _check_type(timeout, ('numeric', ), item_name='timeout')
    assert 0 < timeout

    watchdog = Timer()
    while watchdog.sec() <= timeout:
        stream_list, streamInfos = list_lsl_streams(ignore_markers)
        if len(stream_list) != 0:
            break
    else:
        logger.error('Timeout. No LSL stream found.')
        return None

    logger.info('-- List of servers --')
    for i, stream_name in enumerate(stream_list):
        logger.info(f'{i}: {stream_name}')

    if len(stream_list) == 0:
        logger.error('No LSL stream found on the network.')
    elif len(stream_list) == 1:
        index = 0
    else:
        index = input(
            'Stream index? '
            'Hit enter without index to select the first server.\n>> ')
        if index.strip() == '':
            index = 0
        else:
            index = int(index.strip())

    stream_name = stream_list[index]
    streamInfo = streamInfos[index]
    assert stream_name == streamInfo.name()

    logger.info(f'Selected: {stream_name}')

    return stream_name


def lsl_channel_list(inlet):
    """
    Extract the channels name list from the LSL info.

    Parameters
    ----------
    inlet : pylsl.StreamInlet
        Inlet to extract the channels list from.

    Returns
    -------
    ch_list : list
        List of channels name ``[name1, name2, ... ]``.
    """
    _check_type(inlet, (pylsl.StreamInlet, ), item_name='inlet')

    xml_str = inlet.info().as_xml()
    root = ET.fromstring(xml_str)

    ch_list = []
    for elt in root.iter('channel'):
        ch_list.append(elt.find('label').text)

    return ch_list



from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import QRect, QSize
from PyQt5.QtWidgets import (QWidget, QCheckBox, QLabel, QComboBox,
                             QPushButton, QSpinBox, QDoubleSpinBox,
                             QFrame, QLineEdit, QTableWidget, QStatusBar)

from ..scope._scope import _BUFFER_DURATION


class UI_MainWindow(object):
    def __init__(self, MainWindow):
        self.load_ui(MainWindow)

    def load_ui(self, MainWindow):
        # Set Main window and main widget
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(531, 693)
        MainWindow.setWindowTitle("Scope settings")
        self.MainWidget = QWidget(MainWindow)
        self.MainWidget.setObjectName("MainWidget")
        MainWindow.setCentralWidget(self.MainWidget)

        # Y-Scale of the signal
        self.comboBox_signal_yRange = QComboBox(self.MainWidget)
        self.comboBox_signal_yRange.setGeometry(QRect(10, 20, 85, 27))
        self.comboBox_signal_yRange.setObjectName("comboBox_signal_yRange")

        self.label_signal_yRange = QLabel(self.MainWidget)
        self.label_signal_yRange.setGeometry(QRect(10, 60, 67, 17))
        self.label_signal_yRange.setObjectName("label_signal_yRange")
        self.label_signal_yRange.setText("Time (s)")

        # X-Scale of the signal
        self.spinBox_signal_xRange = QSpinBox(self.MainWidget) # Integers
        self.spinBox_signal_xRange.setGeometry(QRect(10, 80, 85, 27))
        self.spinBox_signal_xRange.setMinimum(1)
        self.spinBox_signal_xRange.setMaximum(int(_BUFFER_DURATION))
        self.spinBox_signal_xRange.setProperty("value", 10) # Default 10s
        self.spinBox_signal_xRange.setObjectName("spinBox_signal_xRange")

        self.label_signal_xRange = QLabel(self.MainWidget)
        self.label_signal_xRange.setGeometry(QRect(10, 0, 67, 17))
        self.label_signal_xRange.setObjectName("label_signal_xRange")
        self.label_signal_xRange.setText("Scale")

        # CAR
        self.checkBox_car = QCheckBox(self.MainWidget)
        self.checkBox_car.setGeometry(QRect(120, 0, 97, 22))
        self.checkBox_car.setObjectName("checkBox_car")
        self.checkBox_car.setText("CAR Filter")

        # BP Filter
        self.checkBox_bandpass = QCheckBox(self.MainWidget)
        self.checkBox_bandpass.setGeometry(QRect(120, 20, 141, 22))
        self.checkBox_bandpass.setObjectName("checkBox_bandpass")
        self.checkBox_bandpass.setText("Bandpass Filter")

        # Set bandpass 'LOW'
        self.doubleSpinBox_bandpass_low = QDoubleSpinBox(self.MainWidget) # Floats
        self.doubleSpinBox_bandpass_low.setGeometry(QRect(250, 17, 69, 27))
        self.doubleSpinBox_bandpass_low.setMinimum(0.1)
        self.doubleSpinBox_bandpass_low.setMaximum(1000.0)
        self.doubleSpinBox_bandpass_low.setSingleStep(1.0)
        self.doubleSpinBox_bandpass_low.setProperty("value", 1.0) # Default 1Hz
        self.doubleSpinBox_bandpass_low.setObjectName("doubleSpinBox_bandpass_low")

        # Set bandpass 'HIGH'
        self.doubleSpinBox_bandpass_high = QDoubleSpinBox(self.MainWidget)
        self.doubleSpinBox_bandpass_high.setGeometry(QRect(330, 17, 69, 27))
        self.doubleSpinBox_bandpass_high.setMinimum(1.0)
        self.doubleSpinBox_bandpass_high.setMaximum(1000.0)
        self.doubleSpinBox_bandpass_high.setSingleStep(1.0)
        self.doubleSpinBox_bandpass_high.setProperty("value", 40.0) # Default 40Hz
        self.doubleSpinBox_bandpass_high.setObjectName("doubleSpinBox_bandpass_high")

        # Show LPT events
        self.checkBox_show_LPT_trigger_events = QCheckBox(self.MainWidget)
        self.checkBox_show_LPT_trigger_events.setGeometry(QRect(120, 80, 151, 22))
        self.checkBox_show_LPT_trigger_events.setObjectName("checkBox_show_LPT_trigger_events")
        self.checkBox_show_LPT_trigger_events.setText("Show LPT events")

        # Lines
        self.line1 = QFrame(self.MainWidget)
        self.line1.setGeometry(QRect(100, -1, 21, 133))
        self.line1.setFrameShape(QFrame.VLine)
        self.line1.setFrameShadow(QFrame.Sunken)
        self.line1.setObjectName("line1")

        self.line2 = QFrame(self.MainWidget)
        self.line2.setGeometry(QRect(110, 45, 421, 21))
        self.line2.setFrameShape(QFrame.HLine)
        self.line2.setFrameShadow(QFrame.Sunken)
        self.line2.setObjectName("line2")

        self.line3 = QFrame(self.MainWidget)
        self.line3.setGeometry(QRect(290, 57, 21, 75))
        self.line3.setFrameShape(QFrame.VLine)
        self.line3.setFrameShadow(QFrame.Sunken)
        self.line3.setObjectName("line3")

        # Start recording
        self.pushButton_start_recording = QPushButton(self.MainWidget)
        self.pushButton_start_recording.setGeometry(QRect(390, 60, 61, 31))
        self.pushButton_start_recording.setObjectName("pushButton_start_recording")
        self.pushButton_start_recording.setText("REC")
        self.pushButton_start_recording.setEnabled(False)

        # Stop recording
        self.pushButton_stop_recording = QPushButton(self.MainWidget)
        self.pushButton_stop_recording.setGeometry(QRect(460, 60, 61, 31))
        self.pushButton_stop_recording.setObjectName("pushButton_stop_recording")
        self.pushButton_stop_recording.setText("Stop REC")
        self.pushButton_stop_recording.setEnabled(False)

        # Set recording directory
        self.pushButton_set_recording_dir = QPushButton(self.MainWidget)
        self.pushButton_set_recording_dir.setGeometry(QRect(310, 60, 71, 31))
        self.pushButton_set_recording_dir.setObjectName("pushButton_set_recording_dir")
        self.pushButton_set_recording_dir.setText("REC Dir")

        # Edit Line for the recording directory
        self.lineEdit_recording_dir = QLineEdit(self.MainWidget)
        self.lineEdit_recording_dir.setGeometry(QRect(310, 100, 211, 27))
        self.lineEdit_recording_dir.setObjectName("lineEdit_recording_dir")

        # Table of channels
        self.table_channels = QTableWidget(self.MainWidget)
        self.table_channels.setGeometry(QRect(4, 131, 525, 503))
        self.table_channels.setMaximumSize(QSize(529, 16777215))
        self.table_channels.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.table_channels.setTextElideMode(QtCore.Qt.ElideMiddle)
        self.table_channels.setShowGrid(False)
        self.table_channels.setObjectName("table_channels")
        self.table_channels.setColumnCount(0)
        self.table_channels.setRowCount(0)
        self.table_channels.horizontalHeader().setVisible(False)
        self.table_channels.horizontalHeader().setHighlightSections(False)
        self.table_channels.verticalHeader().setVisible(False)
        self.table_channels.verticalHeader().setHighlightSections(False)

        # Bottom status Bar
        self.statusBar = QStatusBar(MainWindow)
        self.statusBar.setObjectName("statusBar")
        MainWindow.setStatusBar(self.statusBar)

        # Connect
        # QtCore.QMetaObject.connectSlotsByName(MainWindow)



import pytest

from bsl.datasets._fetching import fetch_file
from bsl.utils._tests import requires_good_network


URL = 'https://raw.githubusercontent.com/bsl-tools/bsl/master/LICENSE'
HASH = {'md5': '1803fa9c2c3ce8cb06b4861d75310742',
        'sha1': 'b386b371ce94933e63ced1052aa72a60da5485ff'}
WRONG_HASH = {'md5': '12345678912345678912345678912345',
              'sha1': '1234567891234567891234567891234567891234'}


@requires_good_network
@pytest.mark.parametrize('hash_type', ('md5', 'sha1'))
def test_fetch_file(tmp_path, hash_type):
    """Test file download."""
    fetch_file(URL, tmp_path / 'README1.md',
               hash_=HASH[hash_type], hash_type=hash_type)
    # Test wrong hash value
    with pytest.raises(RuntimeError, match='Hash mismatch for downloaded'):
        fetch_file(URL, tmp_path / 'README2.md',
                   hash_=WRONG_HASH[hash_type], hash_type=hash_type)


def test_fetch_file_invalid_arg(tmp_path):
    """Test invalid URL and hash/hash types."""
    # Test invalid URL
    with pytest.raises(NotImplementedError, match='Cannot use scheme'):
        fetch_file('not an address', tmp_path / 'test.md')
    # Test invalid hash value
    with pytest.raises(ValueError, match='Bad hash value given'):
        fetch_file(URL, tmp_path / 'test.md', hash_='not an hash')
        fetch_file(URL, tmp_path / 'test.md', hash_=123456789)
    # Test invalid hash type
    with pytest.raises(ValueError, match='Unsupported hash type'):
        fetch_file(URL, tmp_path / 'README3.md',
                   hash_=HASH['md5'], hash_type='...')
        fetch_file(URL, tmp_path / 'README3.md',
                   hash_=HASH['md5'], hash_type=123)



"""Datasets utilities. Inspired from MNE."""

from . import eeg_auditory_stimuli  # noqa: F401
from . import eeg_resting_state  # noqa: F401
from . import eeg_resting_state_short  # noqa: F401
from . import trigger_def  # noqa: F401


__all__ = ['eeg_auditory_stimuli',
           'eeg_resting_state',
           'eeg_resting_state_short',
           'trigger_def']


def _download_all_datasets():
    """
    Download all the datasets.
    """
    eeg_auditory_stimuli.data_path()
    eeg_resting_state.data_path()
    eeg_resting_state_short.data_path()
    trigger_def.data_path()



"""
BSL provides a real-time brain signal streaming framework.
"""

from ._version import __version__  # noqa: F401
from .stream_receiver import StreamReceiver  # noqa: F401
from .stream_recorder import StreamRecorder  # noqa: F401
from .stream_player import StreamPlayer  # noqa: F401
from .stream_viewer import StreamViewer  # noqa: F401
from . import datasets  # noqa: F401
from .utils._logs import (logger, set_log_level,  # noqa: F401
                          set_handler_log_level, add_stream_handler,
                          add_file_handler)



import time
import multiprocessing as mp

import mne
import pytest

from bsl import StreamPlayer, logger, set_log_level
from bsl.datasets import (eeg_auditory_stimuli, eeg_resting_state,
                          eeg_resting_state_short, trigger_def)
from bsl.externals import pylsl
from bsl.triggers import TriggerDef
from bsl.utils._tests import (requires_eeg_auditory_stimuli_dataset,
                              requires_eeg_resting_state_dataset,
                              requires_eeg_resting_state_short_dataset,
                              requires_trigger_def_dataset)


set_log_level('INFO')
logger.propagate = True


@requires_eeg_resting_state_dataset
def test_stream_player(caplog):
    """Test stream player default capabilities."""
    stream_name = 'StreamPlayer'
    fif_file = eeg_resting_state.data_path()
    raw = mne.io.read_raw_fif(fif_file, preload=True)

    # Test start
    caplog.clear()
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file)
    sp.start()
    assert 'Streaming started.' in caplog.text

    # Test stream is in resolved streams
    streams = pylsl.resolve_streams()
    assert stream_name in [stream.name() for stream in streams]

    # Test that data is being streamed
    idx = [stream.name() for stream in streams].index(stream_name)
    inlet = pylsl.StreamInlet(streams[idx], max_buflen=int(raw.info['sfreq']))
    inlet.open_stream()
    time.sleep(0.1)
    chunk, tslist = inlet.pull_chunk(
        timeout=0.0, max_samples=int(raw.info['sfreq']))
    assert len(chunk) == len(tslist)
    assert 0 < len(chunk) < int(raw.info['sfreq'])
    time.sleep(1)
    chunk, tslist = inlet.pull_chunk(
        timeout=0.0, max_samples=int(raw.info['sfreq']))
    assert len(chunk) == len(tslist) == int(raw.info['sfreq'])

    # Test stop
    caplog.clear()
    sp.stop()
    assert ('Waiting for StreamPlayer %s process to finish.'
            % 'StreamPlayer') in caplog.text
    assert 'Streaming finished.' in caplog.text
    assert sp.process is None

    # Test restart/stop
    caplog.clear()
    sp.start()
    assert 'Streaming started.' in caplog.text
    streams = pylsl.resolve_streams()
    assert sp.stream_name in [stream.name() for stream in streams]
    sp.stop()
    assert ('Waiting for StreamPlayer %s process to finish.'
            % 'StreamPlayer') in caplog.text
    assert sp.process is None

    # Test stop when not started
    caplog.clear()
    sp.stop()
    assert 'StreamPlayer was not started. Skipping.' in caplog.text

    # Test context manager
    caplog.clear()
    with StreamPlayer(stream_name=stream_name, fif_file=fif_file):
        assert 'Streaming started.' in caplog.text
        streams = pylsl.resolve_streams()
        assert stream_name in [stream.name() for stream in streams]
        time.sleep(0.5)
    assert ('Waiting for StreamPlayer %s process to finish.'
            % 'StreamPlayer') in caplog.text
    assert 'Streaming finished.' in caplog.text


@requires_eeg_resting_state_short_dataset
def test_arg_repeat():
    """Test stream player replay capabilities."""
    sp = StreamPlayer(stream_name='StreamPlayer',
                      fif_file=eeg_resting_state_short.data_path(),
                      repeat=2)
    sp.start()
    assert sp.state.value == 1
    time.sleep(2.1)
    assert sp.state.value == 1
    time.sleep(2.1)
    assert sp.state.value == 0


@requires_eeg_auditory_stimuli_dataset
def test_arg_trigger_def():
    """Test stream player trigger display capabilities."""
    stream_name = 'StreamPlayer'
    fif_file = eeg_auditory_stimuli.data_path()

    # Without TriggerDef
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      trigger_def=None)
    sp.start()
    time.sleep(2)
    sp.stop()

    # With created TriggerDef
    trigger_def = TriggerDef()
    trigger_def.add('rest', 1)
    trigger_def.add('stim', 4)
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      trigger_def=trigger_def)
    sp.start()
    time.sleep(2)
    sp.stop()

    # With creatred TriggerDef missing an event
    trigger_def = TriggerDef()
    trigger_def.add('rest', 1)
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      trigger_def=trigger_def)
    sp.start()
    time.sleep(2)
    sp.stop()


@requires_eeg_resting_state_dataset
def test_arg_high_resolution():
    """Test stream player high-resolution capabilities."""
    sp = StreamPlayer(stream_name='StreamPlayer',
                      fif_file=eeg_resting_state.data_path(),
                      high_resolution=True)
    assert sp.high_resolution is True
    sp.start()
    assert sp.high_resolution is True
    time.sleep(0.1)
    sp.stop()
    assert sp.high_resolution is True


@requires_eeg_resting_state_dataset
def test_properties():
    """Test the StreamPlayer properties."""
    sp = StreamPlayer(stream_name='StreamPlayer',
                      fif_file=eeg_resting_state.data_path(),
                      repeat=float('inf'),
                      trigger_def=None,
                      chunk_size=16,
                      high_resolution=False)

    # Check getters
    assert sp.stream_name == 'StreamPlayer'
    assert sp.fif_file == eeg_resting_state.data_path()
    assert sp.repeat == float('inf')
    assert sp.trigger_def is None
    assert sp.chunk_size == 16
    assert sp.high_resolution is False

    # Check setters
    with pytest.raises(AttributeError, match="can't set attribute"):
        sp.stream_name = 'new name'
    with pytest.raises(AttributeError, match="can't set attribute"):
        sp.fif_file = 'new fif file'
    with pytest.raises(AttributeError, match="can't set attribute"):
        sp.repeat = 10
    with pytest.raises(AttributeError, match="can't set attribute"):
        sp.trigger_def = TriggerDef()
    with pytest.raises(AttributeError, match="can't set attribute"):
        sp.chunk_size = 8
    with pytest.raises(AttributeError, match="can't set attribute"):
        sp.high_resolution = True

    # Process and state
    assert sp.process is None
    assert isinstance(sp.state, mp.sharedctypes.Synchronized)
    assert sp.state.value == 0

    with pytest.raises(AttributeError, match="can't set attribute"):
        sp.process = 5
    with pytest.raises(AttributeError, match="can't set attribute"):
        sp.state = 5

    sp.start()
    assert isinstance(sp.process, mp.context.Process)
    assert sp.state.value == 1

    sp.stop()
    assert sp.process is None
    assert sp.state.value == 0


@requires_eeg_resting_state_dataset
def test_checker_arguments():
    """Test the argument error checking."""
    stream_name = 'StreamPlayer'
    fif_file = eeg_resting_state.data_path()

    with pytest.raises(TypeError,
                       match="'high_resolution' must be an instance"):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     high_resolution=1)


def test_checker_fif_file():
    """Test the checker for argument fif file."""
    stream_name = 'StreamPlayer'

    with pytest.raises(TypeError,
                       match="'fif_file' must be an instance"):
        StreamPlayer(stream_name=stream_name, fif_file=5)
    with pytest.raises(FileNotFoundError, match='fname does not exist'):
        StreamPlayer(stream_name=stream_name, fif_file='101-file')


@requires_eeg_resting_state_dataset
def test_checker_repeat():
    """Test the checker for argument repeat."""
    stream_name = 'StreamPlayer'
    fif_file = eeg_resting_state.data_path()

    # Default
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      repeat=float('inf'))
    assert sp.repeat == float('inf')

    # Positive integer
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      repeat=101)
    assert sp.repeat == 101

    # Positive float
    with pytest.raises(TypeError, match="'repeat' must be an instance"):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     repeat=101.)

    # Negative integer
    with pytest.raises(ValueError, match='Argument repeat must be a strictly '
                                         'positive integer. Provided: -101'):
        sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                          repeat=-101)

    # Not integer
    with pytest.raises(TypeError, match="'repeat' must be an instance"):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     repeat=[1, 0, 1])


@requires_trigger_def_dataset
@requires_eeg_resting_state_dataset
def test_checker_trigger_def():
    """Test the checker for argument trigger_def."""
    stream_name = 'StreamPlayer'
    fif_file = eeg_resting_state.data_path()
    trigger_file = trigger_def.data_path()

    # Default
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      trigger_def=None)
    assert sp.trigger_def is None

    # TriggerDef instance
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      trigger_def=TriggerDef(trigger_file=None))
    assert isinstance(sp.trigger_def, TriggerDef)
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      trigger_def=TriggerDef(trigger_file=trigger_file))
    assert isinstance(sp.trigger_def, TriggerDef)

    # Path to a non-existing file
    with pytest.raises(ValueError, match='Argument trigger_def is a path that '
                                         'does not exist. Provided: 101-path'):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     trigger_def='101-path')

    # Invalid type
    with pytest.raises(TypeError, match="'trigger_def' must be an instance"):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     trigger_def=101)


@requires_eeg_resting_state_dataset
def test_checker_chunk_size(caplog):
    """Test the checker for argument chunk_size."""
    stream_name = 'StreamPlayer'
    fif_file = eeg_resting_state.data_path()

    # Default
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      chunk_size=16)
    assert sp.chunk_size == 16

    # Positive integer
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      chunk_size=32)
    assert sp.chunk_size == 32

    # Positive float
    with pytest.raises(TypeError, match="'chunk_size' must be an instance"):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     chunk_size=32.)

    # Positive non-usual integer
    caplog.clear()
    sp = StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                      chunk_size=8)
    assert ('The chunk size %s is different from the usual '
            'values 16 or 32.' % 8) in caplog.text
    assert sp.chunk_size == 8

    # Negative number
    with pytest.raises(ValueError,
                       match='Argument chunk_size must be a strictly positive '
                             'integer. Provided: -101'):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     chunk_size=-101)

    # Invalid type
    with pytest.raises(TypeError, match="'chunk_size' must be an instance"):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     chunk_size=[1, 0, 1])

    # Infinite
    with pytest.raises(TypeError, match="'chunk_size' must be an instance"):
        StreamPlayer(stream_name=stream_name, fif_file=fif_file,
                     chunk_size=float('inf'))


@requires_eeg_resting_state_dataset
def test_representation():
    """Test the representation method."""
    fif_file = eeg_resting_state.data_path()
    sp = StreamPlayer(stream_name='StreamPlayer', fif_file=fif_file)
    expected = f'<Player: StreamPlayer | OFF | {fif_file}>'
    assert sp.__repr__() == expected
    sp.start()
    expected = f'<Player: StreamPlayer | ON | {fif_file}>'
    assert sp.__repr__() == expected
    sp.stop()
    expected = f'<Player: StreamPlayer | OFF | {fif_file}>'
    assert sp.__repr__() == expected



from bsl.datasets import trigger_def
from bsl.datasets._fetching import _hashfunc
from bsl.utils._tests import requires_good_network


@requires_good_network
def test_data_path():
    """Test that the path exist and match the desired dataset."""
    path = trigger_def.data_path()
    assert path.exists()
    assert _hashfunc(path, hash_type='md5') == trigger_def.MD5
    assert _hashfunc(path, hash_type='sha1') == trigger_def.SHA1



"""
Module for signal acquisition.

Acquire signals from LSL streams. Used by other modules such as StreamViewer
and StreamRecorder.
"""

from ._stream import StreamEEG, StreamMarker  # noqa: F401
from .stream_receiver import StreamReceiver  # noqa: F401


__all__ = ['StreamReceiver', 'StreamEEG', 'StreamMarker']
